#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import re
import numpy as np
import logging
from datetime import datetime
from ase.data import chemical_symbols

from nomad.units import ureg
from nomad.parsing.parser import FairdiParser
from nomad.parsing.file_parser.text_parser import TextParser, Quantity, DataTextParser
from nomad.datamodel.metainfo.common_dft import Run, Method, MethodBasisSet,\
    BasisSetCellDependent, XCFunctionals, System, SingleConfigurationCalculation,\
    ScfIteration, Dos, Eigenvalues, SamplingMethod
from .metainfo import m_env
from .metainfo.abinit_autogenerated import x_abinit_section_dataset, x_abinit_section_input


ABINIT_NATIVE_IXC = {
    0: [{}],
    1: [{'XC_functional_name': 'LDA_XC_TETER93'}],
    2: [{'XC_functional_name': 'LDA_X'}, {'XC_functional_name': 'LDA_C_PZ'}],
    # 3 - LDA, old Teter rational polynomial parametrization (4/91)
    4: [{'XC_functional_name': 'LDA_X'}, {'XC_functional_name': 'LDA_C_WIGNER'}],
    5: [{'XC_functional_name': 'LDA_X'}, {'XC_functional_name': 'LDA_C_HL'}],
    6: [{'XC_functional_name': 'LDA_X'}, {'XC_functional_name': 'LDA_C_XALPHA'}],
    7: [{'XC_functional_name': 'LDA_X'}, {'XC_functional_name': 'LDA_C_PW'}],
    # 8 - x-only part of the Perdew-Wang 92 functional
    # 9 - x- and RPA correlation part of the Perdew-Wang 92 functional
    # 10 - non-existent
    11: [{'XC_functional_name': 'GGA_X_PBE'}, {'XC_functional_name': 'GGA_C_PBE'}],
    12: [{'XC_functional_name': 'GGA_X_PBE'}],
    13: [{'XC_functional_name': 'GGA_X_LB'}, {'XC_functional_name': 'LDA_C_PW'}],
    14: [{'XC_functional_name': 'GGA_X_PBE_R'}, {'XC_functional_name': '?'}],
    15: [{'XC_functional_name': 'GGA_X_RPBE'}, {'XC_functional_name': '?'}],
    16: [{'XC_functional_name': 'GGA_XC_HCTH_93'}],
    17: [{'XC_functional_name': 'GGA_XC_HCTH_120'}],
    18: [{'XC_functional_name': 'GGA_X_B88'}, {'XC_functional_name': 'GGA_C_LYP'}],
    19: [{'XC_functional_name': 'GGA_X_B88'}, {'XC_functional_name': 'GGA_C_P86'}],
    # 20 - Fermi-Amaldi xc ( -1/N Hartree energy, where N is the number of electrons per cell;
    #      G=0 is not taken into account however), for TDDFT tests.
    # 21 - same as 20, except that the xc-kernel is the LDA (ixc=1) one, for TDDFT tests.
    # 22 - same as 20, except that the xc-kernel is the Burke-Petersilka-Gross hybrid, for
    #      TDDFT tests.
    23: [{'XC_functional_name': 'GGA_X_WC'}, {'XC_functional_name': '?'}],
    24: [{'XC_functional_name': 'GGA_X_C09X'}, {'XC_functional_name': '?'}],
    # 25 - non-existent
    26: [{'XC_functional_name': 'GGA_XC_HCTH_147'}],
    27: [{'XC_functional_name': 'GGA_XC_HCTH_407'}],
    28: [{'XC_functional_name': 'GGA_X_OPTX'}, {'XC_functional_name': 'GGA_C_LYP'}],
    40: [{'XC_functional_name': 'HF_X'}],
    41: [{'XC_functional_name': 'HYB_GGA_XC_PBEH'}],
    42: [{'XC_functional_name': 'HYB_GGA_XC_PBE0_13'}]}

ABINIT_LIBXC_IXC = {
    1: {"XC_functional_name": "LDA_X"},
    2: {"XC_functional_name": "LDA_C_WIGNER"},
    3: {"XC_functional_name": "LDA_C_RPA"},
    4: {"XC_functional_name": "LDA_C_HL"},
    5: {"XC_functional_name": "LDA_C_GL"},
    6: {"XC_functional_name": "LDA_C_XALPHA"},
    7: {"XC_functional_name": "LDA_C_VWN"},
    8: {"XC_functional_name": "LDA_C_VWN_RPA"},
    9: {"XC_functional_name": "LDA_C_PZ"},
    10: {"XC_functional_name": "LDA_C_PZ_MOD"},
    11: {"XC_functional_name": "LDA_C_OB_PZ"},
    12: {"XC_functional_name": "LDA_C_PW"},
    13: {"XC_functional_name": "LDA_C_PW_MOD"},
    14: {"XC_functional_name": "LDA_C_OB_PW"},
    15: {"XC_functional_name": "LDA_C_2D_AMGB"},
    16: {"XC_functional_name": "LDA_C_2D_PRM"},
    17: {"XC_functional_name": "LDA_C_vBH"},
    18: {"XC_functional_name": "LDA_C_1D_CSC"},
    19: {"XC_functional_name": "LDA_X_2D"},
    20: {"XC_functional_name": "LDA_XC_TETER93"},
    21: {"XC_functional_name": "LDA_X_1D"},
    22: {"XC_functional_name": "LDA_C_ML1"},
    23: {"XC_functional_name": "LDA_C_ML2"},
    24: {"XC_functional_name": "LDA_C_GOMBAS"},
    25: {"XC_functional_name": "LDA_C_PW_RPA"},
    26: {"XC_functional_name": "LDA_C_1D_LOOS"},
    27: {"XC_functional_name": "LDA_C_RC04"},
    28: {"XC_functional_name": "LDA_C_VWN_1"},
    29: {"XC_functional_name": "LDA_C_VWN_2"},
    30: {"XC_functional_name": "LDA_C_VWN_3"},
    31: {"XC_functional_name": "LDA_C_VWN_4"},
    32: {"XC_functional_name": "GGA_X_GAM"},
    33: {"XC_functional_name": "GGA_C_GAM"},
    34: {"XC_functional_name": "GGA_X_HCTH_A"},
    35: {"XC_functional_name": "GGA_X_EV93"},
    36: {"XC_functional_name": "HYB_MGGA_X_DLDF"},
    37: {"XC_functional_name": "MGGA_C_DLDF"},
    38: {"XC_functional_name": "GGA_X_BCGP"},
    39: {"XC_functional_name": "GGA_C_BCGP"},
    40: {"XC_functional_name": "GGA_X_LAMBDA_OC2_N"},
    41: {"XC_functional_name": "GGA_X_B86_R"},
    42: {"XC_functional_name": "MGGA_XC_ZLP"},
    43: {"XC_functional_name": "LDA_XC_ZLP"},
    44: {"XC_functional_name": "GGA_X_LAMBDA_CH_N"},
    45: {"XC_functional_name": "GGA_X_LAMBDA_LO_N"},
    46: {"XC_functional_name": "GGA_X_HJS_B88_V2"},
    47: {"XC_functional_name": "GGA_C_Q2D"},
    48: {"XC_functional_name": "GGA_X_Q2D"},
    49: {"XC_functional_name": "GGA_X_PBE_MOL"},
    50: {"XC_functional_name": "LDA_K_TF"},
    51: {"XC_functional_name": "LDA_K_LP"},
    52: {"XC_functional_name": "GGA_K_TFVW"},
    53: {"XC_functional_name": "GGA_K_REVAPBEINT"},
    54: {"XC_functional_name": "GGA_K_APBEINT"},
    55: {"XC_functional_name": "GGA_K_REVAPBE"},
    56: {"XC_functional_name": "GGA_X_AK13"},
    57: {"XC_functional_name": "GGA_K_MEYER"},
    58: {"XC_functional_name": "GGA_X_LV_RPW86"},
    59: {"XC_functional_name": "GGA_X_PBE_TCA"},
    60: {"XC_functional_name": "GGA_X_PBEINT"},
    61: {"XC_functional_name": "GGA_C_ZPBEINT"},
    62: {"XC_functional_name": "GGA_C_PBEINT"},
    63: {"XC_functional_name": "GGA_C_ZPBESOL"},
    64: {"XC_functional_name": "MGGA_XC_OTPSS_D"},
    65: {"XC_functional_name": "GGA_XC_OPBE_D"},
    66: {"XC_functional_name": "GGA_XC_OPWLYP_D"},
    67: {"XC_functional_name": "GGA_XC_OBLYP_D"},
    68: {"XC_functional_name": "GGA_X_VMT84_GE"},
    69: {"XC_functional_name": "GGA_X_VMT84_PBE"},
    70: {"XC_functional_name": "GGA_X_VMT_GE"},
    71: {"XC_functional_name": "GGA_X_VMT_PBE"},
    72: {"XC_functional_name": "MGGA_C_CS"},
    73: {"XC_functional_name": "MGGA_C_MN12_SX"},
    74: {"XC_functional_name": "MGGA_C_MN12_L"},
    75: {"XC_functional_name": "MGGA_C_M11_L"},
    76: {"XC_functional_name": "MGGA_C_M11"},
    77: {"XC_functional_name": "MGGA_C_M08_SO"},
    78: {"XC_functional_name": "MGGA_C_M08_HX"},
    79: {"XC_functional_name": "GGA_C_N12_SX"},
    80: {"XC_functional_name": "GGA_C_N12"},
    81: {"XC_functional_name": "HYB_GGA_X_N12_SX"},
    82: {"XC_functional_name": "GGA_X_N12"},
    83: {"XC_functional_name": "GGA_C_REGTPSS"},
    84: {"XC_functional_name": "GGA_C_OP_XALPHA"},
    85: {"XC_functional_name": "GGA_C_OP_G96"},
    86: {"XC_functional_name": "GGA_C_OP_PBE"},
    87: {"XC_functional_name": "GGA_C_OP_B88"},
    88: {"XC_functional_name": "GGA_C_FT97"},
    89: {"XC_functional_name": "GGA_C_SPBE"},
    90: {"XC_functional_name": "GGA_X_SSB_SW"},
    91: {"XC_functional_name": "GGA_X_SSB"},
    92: {"XC_functional_name": "GGA_X_SSB_D"},
    93: {"XC_functional_name": "GGA_XC_HCTH_407P"},
    94: {"XC_functional_name": "GGA_XC_HCTH_P76"},
    95: {"XC_functional_name": "GGA_XC_HCTH_P14"},
    96: {"XC_functional_name": "GGA_XC_B97_GGA1"},
    97: {"XC_functional_name": "GGA_C_HCTH_A"},
    98: {"XC_functional_name": "GGA_X_BPCCAC"},
    99: {"XC_functional_name": "GGA_C_REVTCA"},
    100: {"XC_functional_name": "GGA_C_TCA"},
    101: {"XC_functional_name": "GGA_X_PBE"},
    102: {"XC_functional_name": "GGA_X_PBE_R"},
    103: {"XC_functional_name": "GGA_X_B86"},
    104: {"XC_functional_name": "GGA_X_HERMAN"},
    105: {"XC_functional_name": "GGA_X_B86_MGC"},
    106: {"XC_functional_name": "GGA_X_B88"},
    107: {"XC_functional_name": "GGA_X_G96"},
    108: {"XC_functional_name": "GGA_X_PW86"},
    109: {"XC_functional_name": "GGA_X_PW91"},
    110: {"XC_functional_name": "GGA_X_OPTX"},
    111: {"XC_functional_name": "GGA_X_DK87_R1"},
    112: {"XC_functional_name": "GGA_X_DK87_R2"},
    113: {"XC_functional_name": "GGA_X_LG93"},
    114: {"XC_functional_name": "GGA_X_FT97_A"},
    115: {"XC_functional_name": "GGA_X_FT97_B"},
    116: {"XC_functional_name": "GGA_X_PBE_SOL"},
    117: {"XC_functional_name": "GGA_X_RPBE"},
    118: {"XC_functional_name": "GGA_X_WC"},
    119: {"XC_functional_name": "GGA_X_MPW91"},
    120: {"XC_functional_name": "GGA_X_AM05"},
    121: {"XC_functional_name": "GGA_X_PBEA"},
    122: {"XC_functional_name": "GGA_X_MPBE"},
    123: {"XC_functional_name": "GGA_X_XPBE"},
    124: {"XC_functional_name": "GGA_X_2D_B86_MGC"},
    125: {"XC_functional_name": "GGA_X_BAYESIAN"},
    126: {"XC_functional_name": "GGA_X_PBE_JSJR"},
    127: {"XC_functional_name": "GGA_X_2D_B88"},
    128: {"XC_functional_name": "GGA_X_2D_B86"},
    129: {"XC_functional_name": "GGA_X_2D_PBE"},
    130: {"XC_functional_name": "GGA_C_PBE"},
    131: {"XC_functional_name": "GGA_C_LYP"},
    132: {"XC_functional_name": "GGA_C_P86"},
    133: {"XC_functional_name": "GGA_C_PBE_SOL"},
    134: {"XC_functional_name": "GGA_C_PW91"},
    135: {"XC_functional_name": "GGA_C_AM05"},
    136: {"XC_functional_name": "GGA_C_XPBE"},
    137: {"XC_functional_name": "GGA_C_LM"},
    138: {"XC_functional_name": "GGA_C_PBE_JRGX"},
    139: {"XC_functional_name": "GGA_X_OPTB88_VDW"},
    140: {"XC_functional_name": "GGA_X_PBEK1_VDW"},
    141: {"XC_functional_name": "GGA_X_OPTPBE_VDW"},
    142: {"XC_functional_name": "GGA_X_RGE2"},
    143: {"XC_functional_name": "GGA_C_RGE2"},
    144: {"XC_functional_name": "GGA_X_RPW86"},
    145: {"XC_functional_name": "GGA_X_KT1"},
    146: {"XC_functional_name": "GGA_XC_KT2"},
    147: {"XC_functional_name": "GGA_C_WL"},
    148: {"XC_functional_name": "GGA_C_WI"},
    149: {"XC_functional_name": "GGA_X_MB88"},
    150: {"XC_functional_name": "GGA_X_SOGGA"},
    151: {"XC_functional_name": "GGA_X_SOGGA11"},
    152: {"XC_functional_name": "GGA_C_SOGGA11"},
    153: {"XC_functional_name": "GGA_C_WI0"},
    154: {"XC_functional_name": "GGA_XC_TH1"},
    155: {"XC_functional_name": "GGA_XC_TH2"},
    156: {"XC_functional_name": "GGA_XC_TH3"},
    157: {"XC_functional_name": "GGA_XC_TH4"},
    158: {"XC_functional_name": "GGA_X_C09X"},
    159: {"XC_functional_name": "GGA_C_SOGGA11_X"},
    160: {"XC_functional_name": "GGA_X_LB"},
    161: {"XC_functional_name": "GGA_XC_HCTH_93"},
    162: {"XC_functional_name": "GGA_XC_HCTH_120"},
    163: {"XC_functional_name": "GGA_XC_HCTH_147"},
    164: {"XC_functional_name": "GGA_XC_HCTH_407"},
    165: {"XC_functional_name": "GGA_XC_EDF1"},
    166: {"XC_functional_name": "GGA_XC_XLYP"},
    167: {"XC_functional_name": "GGA_X_EB88"},
    168: {"XC_functional_name": "GGA_C_PBE_MOL"},
    169: {"XC_functional_name": "HYB_GGA_XC_PBE_MOL0"},
    170: {"XC_functional_name": "GGA_XC_B97_D"},
    171: {"XC_functional_name": "HYB_GGA_XC_PBE_SOL0"},
    172: {"XC_functional_name": "HYB_GGA_XC_PBEB0"},
    173: {"XC_functional_name": "GGA_XC_PBE1W"},
    174: {"XC_functional_name": "GGA_XC_MPWLYP1W"},
    175: {"XC_functional_name": "GGA_XC_PBELYP1W"},
    176: {"XC_functional_name": "HYB_GGA_XC_PBE_MOLB0"},
    177: {"XC_functional_name": "GGA_K_ABSP3"},
    178: {"XC_functional_name": "GGA_K_ABSP4"},
    182: {"XC_functional_name": "GGA_X_LBM"},
    183: {"XC_functional_name": "GGA_X_OL2"},
    184: {"XC_functional_name": "GGA_X_APBE"},
    185: {"XC_functional_name": "GGA_K_APBE"},
    186: {"XC_functional_name": "GGA_C_APBE"},
    187: {"XC_functional_name": "GGA_K_TW1"},
    188: {"XC_functional_name": "GGA_K_TW2"},
    189: {"XC_functional_name": "GGA_K_TW3"},
    190: {"XC_functional_name": "GGA_K_TW4"},
    191: {"XC_functional_name": "GGA_X_HTBS"},
    192: {"XC_functional_name": "GGA_X_AIRY"},
    193: {"XC_functional_name": "GGA_X_LAG"},
    194: {"XC_functional_name": "GGA_XC_MOHLYP"},
    195: {"XC_functional_name": "GGA_XC_MOHLYP2"},
    196: {"XC_functional_name": "GGA_XC_TH_FL"},
    197: {"XC_functional_name": "GGA_XC_TH_FC"},
    198: {"XC_functional_name": "GGA_XC_TH_FCFO"},
    199: {"XC_functional_name": "GGA_XC_TH_FCO"},
    200: {"XC_functional_name": "GGA_C_OPTC"},
    201: {"XC_functional_name": "MGGA_X_LTA"},
    202: {"XC_functional_name": "MGGA_X_TPSS"},
    203: {"XC_functional_name": "MGGA_X_M06_L"},
    204: {"XC_functional_name": "MGGA_X_GVT4"},
    205: {"XC_functional_name": "MGGA_X_TAU_HCTH"},
    206: {"XC_functional_name": "MGGA_X_BR89"},
    207: {"XC_functional_name": "MGGA_X_BJ06"},
    208: {"XC_functional_name": "MGGA_X_TB09"},
    209: {"XC_functional_name": "MGGA_X_RPP09"},
    210: {"XC_functional_name": "MGGA_X_2D_PRHG07"},
    211: {"XC_functional_name": "MGGA_X_2D_PRHG07_PRP10"},
    212: {"XC_functional_name": "MGGA_X_REVTPSS"},
    213: {"XC_functional_name": "MGGA_X_PKZB"},
    214: {"XC_functional_name": "MGGA_X_M05"},
    215: {"XC_functional_name": "MGGA_X_M05_2X"},
    216: {"XC_functional_name": "MGGA_X_M06_HF"},
    217: {"XC_functional_name": "MGGA_X_M06"},
    218: {"XC_functional_name": "MGGA_X_M06_2X"},
    219: {"XC_functional_name": "MGGA_X_M08_HX"},
    220: {"XC_functional_name": "MGGA_X_M08_SO"},
    221: {"XC_functional_name": "MGGA_X_MS0"},
    222: {"XC_functional_name": "MGGA_X_MS1"},
    223: {"XC_functional_name": "MGGA_X_MS2"},
    224: {"XC_functional_name": "HYB_MGGA_X_MS2H"},
    225: {"XC_functional_name": "MGGA_X_M11"},
    226: {"XC_functional_name": "MGGA_X_M11_L"},
    227: {"XC_functional_name": "MGGA_X_MN12_L"},
    229: {"XC_functional_name": "MGGA_C_CC06"},
    230: {"XC_functional_name": "MGGA_X_MK00"},
    231: {"XC_functional_name": "MGGA_C_TPSS"},
    232: {"XC_functional_name": "MGGA_C_VSXC"},
    233: {"XC_functional_name": "MGGA_C_M06_L"},
    234: {"XC_functional_name": "MGGA_C_M06_HF"},
    235: {"XC_functional_name": "MGGA_C_M06"},
    236: {"XC_functional_name": "MGGA_C_M06_2X"},
    237: {"XC_functional_name": "MGGA_C_M05"},
    238: {"XC_functional_name": "MGGA_C_M05_2X"},
    239: {"XC_functional_name": "MGGA_C_PKZB"},
    240: {"XC_functional_name": "MGGA_C_BC95"},
    241: {"XC_functional_name": "MGGA_C_REVTPSS"},
    242: {"XC_functional_name": "MGGA_XC_TPSSLYP1W"},
    243: {"XC_functional_name": "MGGA_X_MK00B"},
    244: {"XC_functional_name": "MGGA_X_BLOC"},
    245: {"XC_functional_name": "MGGA_X_MODTPSS"},
    246: {"XC_functional_name": "GGA_C_PBELOC"},
    247: {"XC_functional_name": "MGGA_C_TPSSLOC"},
    248: {"XC_functional_name": "HYB_MGGA_X_MN12_SX"},
    249: {"XC_functional_name": "MGGA_X_MBEEF"},
    250: {"XC_functional_name": "MGGA_X_MBEEFVDW"},
    254: {"XC_functional_name": "MGGA_XC_B97M_V"},
    255: {"XC_functional_name": "GGA_XC_VV10"},
    257: {"XC_functional_name": "MGGA_X_MVS"},
    258: {"XC_functional_name": "GGA_C_PBEFE"},
    259: {"XC_functional_name": "LDA_XC_KSDT"},
    260: {"XC_functional_name": "MGGA_X_MN15_L"},
    261: {"XC_functional_name": "MGGA_C_MN15_L"},
    262: {"XC_functional_name": "GGA_C_OP_PW91"},
    263: {"XC_functional_name": "MGGA_X_SCAN"},
    264: {"XC_functional_name": "HYB_MGGA_X_SCAN0"},
    265: {"XC_functional_name": "GGA_X_PBEFE"},
    266: {"XC_functional_name": "HYB_GGA_XC_B97_1p"},
    267: {"XC_functional_name": "MGGA_C_SCAN"},
    268: {"XC_functional_name": "HYB_MGGA_X_MN15"},
    269: {"XC_functional_name": "MGGA_C_MN15"},
    270: {"XC_functional_name": "GGA_X_CAP"},
    401: {"XC_functional_name": "HYB_GGA_XC_B3PW91"},
    402: {"XC_functional_name": "HYB_GGA_XC_B3LYP"},
    403: {"XC_functional_name": "HYB_GGA_XC_B3P86"},
    404: {"XC_functional_name": "HYB_GGA_XC_O3LYP"},
    405: {"XC_functional_name": "HYB_GGA_XC_mPW1K"},
    406: {"XC_functional_name": "HYB_GGA_XC_PBEH"},
    407: {"XC_functional_name": "HYB_GGA_XC_B97"},
    408: {"XC_functional_name": "HYB_GGA_XC_B97_1"},
    410: {"XC_functional_name": "HYB_GGA_XC_B97_2"},
    411: {"XC_functional_name": "HYB_GGA_XC_X3LYP"},
    412: {"XC_functional_name": "HYB_GGA_XC_B1WC"},
    413: {"XC_functional_name": "HYB_GGA_XC_B97_K"},
    414: {"XC_functional_name": "HYB_GGA_XC_B97_3"},
    415: {"XC_functional_name": "HYB_GGA_XC_MPW3PW"},
    416: {"XC_functional_name": "HYB_GGA_XC_B1LYP"},
    417: {"XC_functional_name": "HYB_GGA_XC_B1PW91"},
    418: {"XC_functional_name": "HYB_GGA_XC_mPW1PW"},
    419: {"XC_functional_name": "HYB_GGA_XC_MPW3LYP"},
    420: {"XC_functional_name": "HYB_GGA_XC_SB98_1a"},
    421: {"XC_functional_name": "HYB_GGA_XC_SB98_1b"},
    422: {"XC_functional_name": "HYB_GGA_XC_SB98_1c"},
    423: {"XC_functional_name": "HYB_GGA_XC_SB98_2a"},
    424: {"XC_functional_name": "HYB_GGA_XC_SB98_2b"},
    425: {"XC_functional_name": "HYB_GGA_XC_SB98_2c"},
    426: {"XC_functional_name": "HYB_GGA_X_SOGGA11_X"},
    427: {"XC_functional_name": "HYB_GGA_XC_HSE03"},
    428: {"XC_functional_name": "HYB_GGA_XC_HSE06"},
    429: {"XC_functional_name": "HYB_GGA_XC_HJS_PBE"},
    430: {"XC_functional_name": "HYB_GGA_XC_HJS_PBE_SOL"},
    431: {"XC_functional_name": "HYB_GGA_XC_HJS_B88"},
    432: {"XC_functional_name": "HYB_GGA_XC_HJS_B97X"},
    433: {"XC_functional_name": "HYB_GGA_XC_CAM_B3LYP"},
    434: {"XC_functional_name": "HYB_GGA_XC_TUNED_CAM_B3LYP"},
    435: {"XC_functional_name": "HYB_GGA_XC_BHANDH"},
    436: {"XC_functional_name": "HYB_GGA_XC_BHANDHLYP"},
    437: {"XC_functional_name": "HYB_GGA_XC_MB3LYP_RC04"},
    438: {"XC_functional_name": "HYB_MGGA_XC_M05"},
    439: {"XC_functional_name": "HYB_MGGA_XC_M05_2X"},
    440: {"XC_functional_name": "HYB_MGGA_XC_B88B95"},
    441: {"XC_functional_name": "HYB_MGGA_XC_B86B95"},
    442: {"XC_functional_name": "HYB_MGGA_XC_PW86B95"},
    443: {"XC_functional_name": "HYB_MGGA_XC_BB1K"},
    444: {"XC_functional_name": "HYB_MGGA_XC_M06_HF"},
    445: {"XC_functional_name": "HYB_MGGA_XC_MPW1B95"},
    446: {"XC_functional_name": "HYB_MGGA_XC_MPWB1K"},
    447: {"XC_functional_name": "HYB_MGGA_XC_X1B95"},
    448: {"XC_functional_name": "HYB_MGGA_XC_XB1K"},
    449: {"XC_functional_name": "HYB_MGGA_XC_M06"},
    450: {"XC_functional_name": "HYB_MGGA_XC_M06_2X"},
    451: {"XC_functional_name": "HYB_MGGA_XC_PW6B95"},
    452: {"XC_functional_name": "HYB_MGGA_XC_PWB6K"},
    453: {"XC_functional_name": "HYB_GGA_XC_MPWLYP1M"},
    454: {"XC_functional_name": "HYB_GGA_XC_REVB3LYP"},
    455: {"XC_functional_name": "HYB_GGA_XC_CAMY_BLYP"},
    456: {"XC_functional_name": "HYB_GGA_XC_PBE0_13"},
    457: {"XC_functional_name": "HYB_MGGA_XC_TPSSH"},
    458: {"XC_functional_name": "HYB_MGGA_XC_REVTPSSH"},
    459: {"XC_functional_name": "HYB_GGA_XC_B3LYPs"},
    460: {"XC_functional_name": "HYB_MGGA_XC_M08_HX"},
    461: {"XC_functional_name": "HYB_MGGA_XC_M08_SO"},
    462: {"XC_functional_name": "HYB_MGGA_XC_M11"},
    463: {"XC_functional_name": "HYB_GGA_XC_WB97"},
    464: {"XC_functional_name": "HYB_GGA_XC_WB97X"},
    465: {"XC_functional_name": "HYB_GGA_XC_LRC_WPBEH"},
    466: {"XC_functional_name": "HYB_GGA_XC_WB97X_V"},
    467: {"XC_functional_name": "HYB_GGA_XC_LCY_PBE"},
    468: {"XC_functional_name": "HYB_GGA_XC_LCY_BLYP"},
    469: {"XC_functional_name": "HYB_GGA_XC_LC_VV10"},
    470: {"XC_functional_name": "HYB_GGA_XC_CAMY_B3LYP"},
    471: {"XC_functional_name": "HYB_GGA_XC_WB97X_D"},
    472: {"XC_functional_name": "HYB_GGA_XC_HPBEINT"},
    473: {"XC_functional_name": "HYB_GGA_XC_LRC_WPBE"},
    474: {"XC_functional_name": "HYB_MGGA_X_MVSH"},
    475: {"XC_functional_name": "HYB_GGA_XC_B3LYP5"},
    476: {"XC_functional_name": "HYB_GGA_XC_EDF2"},
    477: {"XC_functional_name": "HYB_GGA_XC_CAP0"},
    478: {"XC_functional_name": "HYB_GGA_XC_LC_WPBE"},
    500: {"XC_functional_name": "GGA_K_VW"},
    501: {"XC_functional_name": "GGA_K_GE2"},
    502: {"XC_functional_name": "GGA_K_GOLDEN"},
    503: {"XC_functional_name": "GGA_K_YT65"},
    504: {"XC_functional_name": "GGA_K_BALTIN"},
    505: {"XC_functional_name": "GGA_K_LIEB"},
    506: {"XC_functional_name": "GGA_K_ABSP1"},
    507: {"XC_functional_name": "GGA_K_ABSP2"},
    508: {"XC_functional_name": "GGA_K_GR"},
    509: {"XC_functional_name": "GGA_K_LUDENA"},
    510: {"XC_functional_name": "GGA_K_GP85"},
    511: {"XC_functional_name": "GGA_K_PEARSON"},
    512: {"XC_functional_name": "GGA_K_OL1"},
    513: {"XC_functional_name": "GGA_K_OL2"},
    514: {"XC_functional_name": "GGA_K_FR_B88"},
    515: {"XC_functional_name": "GGA_K_FR_PW86"},
    516: {"XC_functional_name": "GGA_K_DK"},
    517: {"XC_functional_name": "GGA_K_PERDEW"},
    518: {"XC_functional_name": "GGA_K_VSK"},
    519: {"XC_functional_name": "GGA_K_VJKS"},
    520: {"XC_functional_name": "GGA_K_ERNZERHOF"},
    521: {"XC_functional_name": "GGA_K_LC94"},
    522: {"XC_functional_name": "GGA_K_LLP"},
    523: {"XC_functional_name": "GGA_K_THAKKAR"},
    524: {"XC_functional_name": "GGA_X_WPBEH"},
    525: {"XC_functional_name": "GGA_X_HJS_PBE"},
    526: {"XC_functional_name": "GGA_X_HJS_PBE_SOL"},
    527: {"XC_functional_name": "GGA_X_HJS_B88"},
    528: {"XC_functional_name": "GGA_X_HJS_B97X"},
    529: {"XC_functional_name": "GGA_X_ITYH"},
    530: {"XC_functional_name": "GGA_X_SFAT"},
    531: {"XC_functional_name": "HYB_MGGA_XC_WB97M_V"}}


class AbinitOutParser(TextParser):
    def __init__(self):
        self.energy_components = {
            'kinetic': 'Kinetic energy',
            'hartree': 'Hartree energy',
            'xc': 'XC energy',
            'ewald': 'Ewald energy',
            'psp_core': 'PspCore energy',
            'psp_local': 'Loc. psp. energy',
            'psp_nonlocal': 'NL   psp  energy',
            'internal': '>>>>> Internal E',
            'ktentropy': r'\-kT*entropy',
            'total': ' >>>>>>>>> Etotal',
            'band': r'Band energy \(Ha\)'}
        self._input_vars = None
        self._n_datasets = None
        super().__init__(None)

    def init_quantities(self):
        self._quantities = [
            Quantity(
                'program_version', r'\.Version ([\w\.]+) of ABINIT', repeats=False,
                convert=False, flatten=False),
            Quantity(
                'x_abinit_parallel_compilation', r'\.\((\w+) version,', repeats=False,
                convert=False, flatten=False),
            Quantity(
                'program_compilation_host', r'prepared for a ([\w\.]+) computer',
                repeats=False, convert=False, flatten=False),
            Quantity(
                'x_abinit_start_date', r'\.Starting date : ([\w ]+)\.', repeats=False,
                convert=False, flatten=False),
            Quantity(
                'x_abinit_start_time', r'\- \( at (\w+) \)', repeats=False, convert=False,
                flatten=False),
            Quantity(
                'x_abinit_input_file', r'\- input\s*file\s*\-\> ([\w\.]+)', repeats=False,
                convert=False, flatten=False),
            Quantity(
                'x_abinit_output_file', r'\- output\s*file\s*\-\> ([\w\.]+)', repeats=False,
                convert=False, flatten=False),
            Quantity(
                'x_abinit_input_files_root', r'\- root for input\s*files \-\> (\w+)',
                repeats=False, convert=False, flatten=False),
            Quantity(
                'x_abinit_output_files_root', r'\- root for output\s*files \-\> (\w+)',
                repeats=False, convert=False, flatten=False),
            Quantity(
                'x_abinit_total_cpu_time',
                r'\-\s*Total cpu\s*time\s*\(s,m,h\):\s*([\d\.]+)', dtype=float),
            Quantity(
                'x_abinit_total_wallclock_time',
                r'\-\s*Total wall clock time\s*\(s,m,h\):\s*([\d\.]+)', dtype=float),
            Quantity(
                'run_clean_end',
                r'(Calculation completed)', repeats=False, convert=False, flatten=False)]

        self._quantities.append(Quantity(
            'input_vars',
            r'\-outvars: echo values of preprocessed input variables \-+([\s\S]+?)\={10}',
            repeats=False, sub_parser=TextParser(quantities=[Quantity(
                'key_value', r'(\w+)\s*([\d\.\+\-E\s]+)', repeats=True)])))

        def str_to_array(val_in):
            val = val_in.strip().split('\n')
            val = [v.split()[-3:] for v in val]
            return np.array(val, dtype=float)

        def str_to_stress_tensor(val_in):
            val = np.array(val_in.split(), dtype=float)
            stress_tensor = np.zeros((3, 3))
            stress_tensor[0][0] = val[0]
            stress_tensor[2][1] = stress_tensor[1][2] = val[1]
            stress_tensor[1][1] = val[2]
            stress_tensor[2][0] = stress_tensor[0][2] = val[3]
            stress_tensor[2][2] = val[4]
            stress_tensor[1][0] = stress_tensor[0][1] = val[5]
            return stress_tensor * (ureg.hartree / ureg.bohr**3)

        def str_to_eigenvalues(val_in):
            return [float(v) for v in val_in.split() if v[-1].isdecimal()]

        self_consistent = [
            Quantity(
                'energy_total_scf_iteration',
                r'ETOT\s*\d+\s*([\+\-\d\.Ee ]+)\n', repeats=True, dtype=float),
            Quantity(
                'convergence',
                r'At SCF step\s*([0-9]+)\s*'
                r', etot|, forces|vres2\s*=\s*[\-\+\d\.Ee]+?\s*<\s*tolvrs=\s*[\-\+\d\.Ee]+?\s*=>\s*'
                r'([\w ]+)', repeats=False),
        ]

        relaxation = [
            Quantity(
                'stress_tensor',
                r'Cartesian components of stress tensor \(hartree/bohr\^3\)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*',
                repeats=False, str_operation=str_to_stress_tensor, convert=False),
            Quantity(
                'self_consistent',
                r'SELF\-CONSISTENT\-FIELD CONVERGENCE\-*\s*(iter\s*Etot[\s\S]+?)\-*OUTPUT',
                sub_parser=TextParser(quantities=self_consistent), repeats=False),
            Quantity(
                'cartesian_coordinates',
                r'Cartesian coordinates \(xcart\) \[bohr\]\s*([\s\d\.Ee\+\-]+)',
                repeats=False, str_operation=str_to_array, convert=False, unit=ureg.bohr),
            Quantity(
                'cartesian_forces',
                r'Cartesian forces \(fcart\)[\s\S]+?\(free atoms\)\s*([\s\d\.Ee\+\-]+)',
                repeats=False, str_operation=str_to_array, convert=False, unit=ureg.hartree / ureg.bohr),
            Quantity(
                'energy_total',
                r'Total energy \(etotal\) \[Ha\]=\s*([\-\+\d\.Ee]+)', repeats=False,
                dtype=float, unit=ureg.hartree)]

        energy_components = [Quantity(
            key, r'\s*%s\s*=\s*([\d\.E\-\+]+)' % val, repeats=False, dtype=float,
            unit=ureg.hartree) for key, val in self.energy_components.items()]

        results = [
            Quantity(
                'cartesian_coordinates',
                r'\s*cartesian coordinates \(angstrom\) at end:\s*([\s\d\.Ee\+\-]+)',
                repeats=False, str_operation=str_to_array, convert=False, unit=ureg.angstrom),
            Quantity(
                'cartesian_forces',
                r'\s*cartesian forces \(hartree/bohr\) at end\:\s*([\s\d\.Ee\+\-]+)',
                repeats=False, str_operation=str_to_array, convert=False, unit=ureg.hartree / ureg.bohr),
            Quantity(
                'x_abinit_eig_filename',
                r'\s*prteigrs : about to open file\s*(\S+)', repeats=False,
                convert=False),
            Quantity(
                'fermi_energy',
                r'\s*Fermi \(or HOMO\) energy \(hartree\) =\s*([-+0-9.]+)',
                repeats=False, dtype=float, unit=ureg.hartree),
            Quantity(
                'x_abinit_magnetisation',
                r'\s*Magnetisation \(Bohr magneton\)=\s*([-+0-9.eEdD]*)\s*',
                repeats=False, dtype=float, unit=ureg.bohr_magneton),
            Quantity(
                'eigenvalues',
                r'\s*kpt#\s*(\d+), nband=\s*(\d+), wtk=\s*([\d\.]+)\s*, '
                r'kpt=([\d\.\-\+Ee ]+)\s*\(reduced coord\)\s*([\d\.\-\+Ee\s]+)',
                repeats=True, str_operation=str_to_eigenvalues, convert=False),
            Quantity(
                'occupation_numbers',
                r'\s*occupation numbers for kpt#\s*\d+\s*([\d\.\-\+Ee ]+)',
                repeats=True, dtype=float),
            Quantity(
                'energy_total',
                r'Total energy\(eV\)=\s*([\d\.\-\+Ee]+)\s*;',
                repeats=False, dtype=float, unit=ureg.eV),
            Quantity(
                'stress_tensor',
                r'Cartesian components of stress tensor \(hartree/bohr\^3\)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*'
                r'sigma\(\d \d\)\s*=\s*([\-\+\d\.Ee]+)\s*',
                repeats=False, str_operation=str_to_stress_tensor, convert=False)
        ] + energy_components

        self._quantities.append(Quantity(
            'dataset',
            r'==\s*(DATASET\s*\d+\s*==[\s\S]+?)(?:\-Cartesian components of stress tensor \(GPa\)|== END DATASET)',
            repeats=True, sub_parser=TextParser(quantities=[
                Quantity(
                    'x_abinit_dataset_number',
                    r'DATASET\s*(\d+)', dtype=int, repeats=False),
                Quantity(
                    'x_abinit_var_ixc',
                    r'\-\s*ixc\s*=\s*([\d\-]+)', dtype=int, repeats=False),
                Quantity(
                    'x_abinit_vprim',
                    r'R\(\d\)=\s*(\-*[\d\.]+)\s*(\-*[\d\.]+)\s*(\-*[\d\.]+)\s*G\(\d\)=',
                    repeats=True, dtype=float),
                Quantity(
                    'x_abinit_unit_cell_volume',
                    r'Unit cell volume ucvol\s*=\s*([\d\.\+Ee]+)', dtype=float,
                    unit=ureg.bohr ** 3),
                Quantity(
                    'self_consistent',
                    r'\={10}\s*(iter\s*Etot[\s\S]+?)\={10}', repeats=False,
                    sub_parser=TextParser(quantities=self_consistent)),
                Quantity(
                    'relaxation',
                    r'\-\-\- Iteration: \( \d+\/\d+\)([\s\S]+?Total energy \(etotal\) \[Ha\]=\s*[\-\+\d\.Ee]+)',
                    repeats=True, sub_parser=TextParser(quantities=relaxation)),
                Quantity(
                    'results',
                    r'\-+iterations are completed or convergence reached\-+([\s\S]+)'
                    r'(?:(sigma\(\d \d\)\s*=\s*[\+\-\d\.E]+)|={80})', repeats=False,
                    sub_parser=TextParser(quantities=results))])))

    @property
    def n_datasets(self):
        if self._n_datasets is None:
            # we could also simply return the length of dataset but this is safer
            dataset_numbers = [
                d.get('x_abinit_dataset_number', 1) for d in self.get('dataset', [])]
            self._n_datasets = max(dataset_numbers) if dataset_numbers else 1
        return self._n_datasets

    @property
    def input_vars(self):
        if self._input_vars is None:
            # set defaults
            self._input_vars = {key: [1] + [None] * (self.n_datasets - 1) for key in [
                'ntypat', 'npsp', 'nshiftk', 'natrd', 'nsppol', 'nspden', 'nkpt',
                'occopt', 'ixc']}

            key_val = self.get('input_vars', {}).get('key_value', [])
            for i in range(len(key_val)):
                key, n_dataset = re.search(r'(\D+)?(\d*)', key_val[i][0]).groups()
                self._input_vars.setdefault(key, [None] * self.n_datasets)

                m_quantity = m_env.all_definitions_by_name.get(
                    'x_abinit_var_%s' % key, [None])[0]
                if m_quantity is None:
                    continue

                val = np.array(key_val[i][1:], dtype=m_quantity.type)
                if not m_quantity.shape:
                    val = val[0]
                if n_dataset:
                    self._input_vars[key][int(n_dataset) - 1] = val
                else:
                    self._input_vars[key] = [val] * self.n_datasets
        return self._input_vars

    def get_input_var(self, key, n_dataset, default=None):
        val = self.input_vars.get(key)
        if val is None:
            val = [default] * n_dataset
        return val[n_dataset - 1]

    def get_atom_labels(self, n_dataset):
        znucl = self.get_input_var('znucl', n_dataset)
        typat = self.get_input_var('typat', n_dataset)
        if znucl is None or typat is None:
            return

        return [chemical_symbols[int(znucl[n_at - 1])] for n_at in typat]


class AbinitParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/abinit', code_name='ABINIT', code_homepage='https://www.abinit.org/',
            mainfile_contents_re=(r'^\n*\.Version\s*[0-9.]*\s*of ABINIT\s*'))
        self._metainfo_env = m_env
        self.out_parser = AbinitOutParser()
        self.dos_parser = DataTextParser()

    def parse_system(self, n_dataset, section):
        dataset = self.out_parser.get('dataset')[n_dataset]
        sec_run = self.archive.section_run[-1]

        nd = dataset.get('x_abinit_dataset_number')
        atom_labels = self.out_parser.get_atom_labels(nd)
        atom_positions = section.get('cartesian_coordinates')
        if atom_labels is None or atom_positions is None:
            return

        sec_system = sec_run.m_create(System)
        sec_system.atom_labels = atom_labels
        sec_system.atom_positions = atom_positions
        sec_system.configuration_periodic_dimensions = [True, True, True]
        sec_system.number_of_atoms = len(atom_labels)
        lattice_vectors = dataset.get('x_abinit_vprim')
        lattice_vectors = np.eye(3) if lattice_vectors is None else lattice_vectors * ureg.bohr
        sec_system.lattice_vectors = lattice_vectors
        sec_system.simulation_cell = lattice_vectors

        return sec_system

    def parse_scc(self, n_dataset, section):
        dataset = self.out_parser.get('dataset')[n_dataset]
        sec_run = self.archive.section_run[-1]

        energy_total = section.get('energy_total')

        sec_scc = sec_run.m_create(SingleConfigurationCalculation)
        sec_scc.energy_total = energy_total
        forces = section.get('cartesian_forces')
        if forces is not None:
            sec_scc.atom_forces_raw = forces

        stress_tensor = section.get('stress_tensor')
        if stress_tensor is not None:
            sec_scc.stress_tensor = stress_tensor

        nsppol = self.out_parser.get_input_var(
            'nsppol', dataset.get('x_abinit_dataset_number'), 1)
        fermi_energy = section.get('fermi_energy')
        if fermi_energy is not None:
            sec_scc.energy_reference_fermi = ([fermi_energy.magnitude] * nsppol) * fermi_energy.units

        # energy contributions, should be moved to EnergyContributions
        # TODO fix metainfo for energy contributions
        for key in self.out_parser.energy_components.keys():
            energy = section.get(key)
            if energy is not None:
                # no units in metainfo
                setattr(sec_scc, 'x_abinit_energy_%s' % key, energy.to('joule').magnitude)

        return sec_scc

    def parse_method(self, n_dataset):
        dataset = self.out_parser.get('dataset')[n_dataset]
        sec_run = self.archive.section_run[-1]
        smearing_kinds = [
            '', '', '', 'fermi', 'marzari-vanderbilt', 'marzari-vanderbilt',
            'methfessel-paxton', 'gaussian', '']
        nd = dataset.get('x_abinit_dataset_number')
        sec_method = sec_run.m_create(Method)

        sec_method.electronic_structure_method = 'DFT'
        sec_method.stress_tensor_method = 'analytic'
        sec_method.number_of_spin_channels = int(self.out_parser.get_input_var('nsppol', nd, 1))
        sec_method.scf_max_iteration = int(self.out_parser.get_input_var('nstep', nd, 1))

        toldfe = self.out_parser.get_input_var('toldfe', nd)
        if toldfe is not None:
            sec_method.scf_threshold_energy_change = toldfe * ureg.hartree
        # not sure what this is for
        sec_method.self_interaction_correction_method = ''
        occopt = self.out_parser.get_input_var('occopt', nd, 1)
        smearing = smearing_kinds[occopt] if occopt < len(smearing_kinds) else ''
        sec_method.smearing_kind = smearing
        tsmear = self.out_parser.get_input_var('tsmear', nd)
        if tsmear is not None:
            sec_method.smearing_width = (tsmear * ureg.hartree).to('joule').magnitude

        sec_method_basis_set = sec_method.m_create(MethodBasisSet)
        sec_method_basis_set.method_basis_set_kind = 'wavefunction'
        sec_method_basis_set.mapping_section_method_basis_set_cell_associated =\
            sec_run.section_basis_set_cell_dependent[-1]

        ixc = self.out_parser.get_input_var('ixc', nd)
        ixc = dataset.get('x_abinit_var_ixc', 0) if ixc is None else ixc

        if ixc >= 0:
            xc_functionals = ABINIT_NATIVE_IXC.get(ixc, [])
        else:
            xc_functionals = []
            functional1 = -ixc // 1000
            if functional1 > 0:
                xc_functionals.append(ABINIT_LIBXC_IXC.get(functional1))
            functional2 = -ixc - (-ixc // 1000) * 1000
            if functional2 > 0:
                xc_functionals.append(ABINIT_LIBXC_IXC.get(functional2))

        for xc_functional in xc_functionals:
            sec_XC_functionals = sec_method.m_create(XCFunctionals)
            for key, value in sorted(xc_functional.items()):
                setattr(sec_XC_functionals, key, value)

        return sec_method

    def parse_sampling_method(self, n_dataset):
        dataset = self.out_parser.get('dataset')[n_dataset]
        sec_run = self.archive.section_run[-1]
        nd = dataset.get('x_abinit_dataset_number')
        sec_sampling_method = sec_run.m_create(SamplingMethod)

        geometry_optimizations = {
            1: 'viscous_damped_md', 2: 'bfgs', 3: 'bfgs', 4: 'conjugate_gradient',
            5: 'steepest_descent', 7: 'quenched_md', 10: 'dic_bfgs', 11: 'dic_bfgs',
            20: 'diis'}

        ionmov = self.out_parser.get_input_var('ionmov', nd)
        vis = self.out_parser.get_input_var('vis', nd, 0.0)
        if ionmov in [2, 3, 4, 5, 7, 10, 11, 20] or (ionmov == 1 and vis > 0.0):
            sampling_method = 'geometry_optimization'
            sec_sampling_method.geometry_optimization_method = geometry_optimizations[ionmov]
        elif ionmov in [6, 8, 12, 13, 14, 23] or (ionmov == 1 and vis == 0.0):
            sampling_method = 'molecular_dynamics'
        elif ionmov == 9:
            sampling_method = 'langevin_dynamics'
        else:
            sampling_method = ''
        sec_sampling_method.sampling_method = sampling_method

        tolmxf = self.out_parser.get_input_var('tolmxf', nd)
        if tolmxf is not None:
            sec_sampling_method.geometry_optimization_threshold_force = tolmxf * (ureg.hartree / ureg.bohr)

        tolmxde = self.out_parser.get_input_var('tolmxde', nd)
        if tolmxde is not None:
            sec_sampling_method.geometry_optimization_energy_change = tolmxde * ureg.hartree

    def parse_dataset(self, n_dataset):
        dataset = self.out_parser.get('dataset')[n_dataset]
        sec_run = self.archive.section_run[-1]
        sec_basis_set_cell_dependent = sec_run.m_create(BasisSetCellDependent)
        sec_basis_set_cell_dependent.basis_set_cell_dependent_kind = 'plane_waves'

        nd = dataset.get('x_abinit_dataset_number')
        ecut = self.out_parser.get_input_var('ecut', nd)
        if ecut is not None:
            ecut = ecut * ureg.hartree
            sec_basis_set_cell_dependent.basis_set_planewave_cutoff = ecut
            name = 'PW_%s' % ecut.to('rydberg').magnitude
            sec_basis_set_cell_dependent.basis_set_cell_dependent_name = name

        # one section_method per dataset
        sec_method = self.parse_method(n_dataset)

        self.parse_sampling_method(n_dataset)

        def parse_configurations(section):
            sec_system = self.parse_system(n_dataset, section)
            sec_scc = self.parse_scc(n_dataset, section)

            if sec_scc is None:
                return

            if sec_system is not None:
                sec_scc.single_configuration_calculation_to_system_ref = sec_system
            sec_scc.single_configuration_to_calculation_method_ref = sec_method

        def parse_dos():
            file_root = self.out_parser.get('x_abinit_output_files_root')
            if file_root is None:
                file_root = '%s_o' % os.path.basename(self.filepath).rstrip('.out')
            dos_file = os.path.join(self.maindir, '%s_DS%d_DOS' % (file_root, nd))
            if not os.path.isfile(dos_file):
                return

            self.dos_parser.mainfile = dos_file
            dos = self.dos_parser.data
            nsppol = self.out_parser.get_input_var('nsppol', nd, 1)
            dos = np.reshape(dos, (nsppol, len(dos) // nsppol, np.size(dos) // len(dos)))

            sec_scc = sec_run.section_single_configuration_calculation[-1]
            sec_dos = sec_scc.m_create(Dos)
            sec_dos.dos_kind = 'electronic'

            sec_dos.dos_energies = dos.T[0].T[0] * ureg.hartree
            sec_dos.number_of_dos_values = np.shape(dos)[1]

            unit_volume = dataset.get('x_abinit_unit_cell_volume')
            dos_values = dos.T[1].T * (1 / ureg.hartree) * unit_volume.magnitude
            sec_dos.dos_values = dos_values.to('1/joule').magnitude
            integrated_dos = dos.T[2].T * unit_volume.magnitude
            sec_dos.dos_integrated_values = integrated_dos

        def parse_eigenvalues():
            data = dataset.get('results', {}).get('eigenvalues', None)
            if data is None:
                return

            nsppol = self.out_parser.get_input_var('nsppol', nd, 1)
            data = np.reshape(data, (nsppol, len(data) // nsppol, np.size(data) // len(data)))

            nband = int(data.T[1].T[0][0])
            kpts = data.T[3:6].T[0]
            eigs = data.T[6:6 + nband].T

            sec_scc = sec_run.section_single_configuration_calculation[-1]
            sec_eigenvalues = sec_scc.m_create(Eigenvalues)
            sec_eigenvalues.eigenvalues_kind = 'normal'
            sec_eigenvalues.number_of_eigenvalues = nband
            sec_eigenvalues.number_of_eigenvalues_kpoints = len(kpts)
            sec_eigenvalues.eigenvalues_kpoints = kpts
            sec_eigenvalues.eigenvalues_values = eigs * ureg.hartree
            occs = dataset.get('results', {}).get('occupation_numbers')
            if occs is not None:
                occs = np.reshape(occs, (nsppol, len(occs) // nsppol, np.size(occs) // len(occs)))
                sec_eigenvalues.eigenvalues_occupation = occs

        def parse_scf(section):
            if not sec_run.section_single_configuration_calculation:
                return

            sec_scc = sec_run.section_single_configuration_calculation[-1]

            self_consistent = section.get('self_consistent', {})
            if self_consistent.get('convergence', None) in ['converged', 'is converged', 'are converged']:
                sec_scc.single_configuration_calculation_converged = True

            energy_total_scf_iteration = self_consistent.get('energy_total_scf_iteration', [])
            for energy in energy_total_scf_iteration:
                sec_scf_iteration = sec_scc.m_create(ScfIteration)
                sec_scf_iteration.energy_total_scf_iteration = energy[0] * ureg.hartree

        # results of the single point calculation
        results = dataset.get('results')
        parse_configurations(results)
        # add scf iterations
        parse_scf(dataset)
        # add dos and eigenvalues to full-scf calculation
        parse_dos()
        parse_eigenvalues()

        # relaxations
        relaxation = dataset.get('relaxation', [])
        for step in relaxation:
            parse_configurations(step)
            parse_scf(step)

    def parse_var(self):
        # TODO reshape input_var arrays
        sec_run = self.archive.section_run[-1]
        for i in range(len(self.out_parser.get('dataset', []))):
            sec_dataset = sec_run.m_create(x_abinit_section_dataset)
            sec_input = sec_dataset.m_create(x_abinit_section_input)
            for key, val in self.out_parser.input_vars.items():
                if val[i] is None:
                    continue
                setattr(sec_input, 'x_abinit_var_%s' % (key), val[i])

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger
        self.out_parser._input_vars = None
        self.out_parser._n_datasets = None

    def reuse_parser(self, parser):
        self.out_parser.quantities = parser.out_parser.quantities

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging

        self.init_parser()

        sec_run = self.archive.m_create(Run)
        sec_run.program_name = 'ABINIT'
        sec_run.program_basis_set_type = 'plane waves'

        section_run_keys = [
            'program_version', 'x_abinit_parallel_compilation', 'program_compilation_host',
            'x_abinit_start_date', 'x_abinit_start_time', 'x_abinit_input_file',
            'x_abinit_output_file', 'x_abinit_input_files_root', 'x_abinit_output_files_root',
            'x_abinit_total_cpu_time', 'x_abinit_total_wallclock_time']

        for key in section_run_keys:
            val = self.out_parser.get(key)
            if val is None:
                continue
            setattr(sec_run, key, val)

        if self.out_parser.get('run_clean_end') is not None:
            sec_run.run_clean_end = True

        abi_time = datetime.strptime(str("%s %s") % (
            self.out_parser.get('x_abinit_start_date'),
            self.out_parser.get('x_abinit_start_time')), "%a %d %b %Y %Hh%M")
        sec_run.time_run_date_start = (abi_time - datetime(1970, 1, 1)).total_seconds()

        self.parse_var()

        for n in range(len(self.out_parser.get('dataset', []))):
            self.parse_dataset(n)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any

from nomad.normalizing.normalizer import Normalizer
from nomad.datamodel.encyclopedia import (
    EncyclopediaMetadata,
    Material,
    Method,
    Properties,
    Calculation,
)
from nomad.normalizing.encyclopedia.context import Context
from nomad.normalizing.encyclopedia.material import MaterialBulkNormalizer, Material2DNormalizer, Material1DNormalizer
from nomad.normalizing.encyclopedia.method import MethodDFTNormalizer, MethodGWNormalizer
from nomad.normalizing.encyclopedia.properties import PropertiesNormalizer
from nomad import config


class EncyclopediaNormalizer(Normalizer):
    """
    This normalizer emulates the functionality of the old Encyclopedia backend.
    The data used by the encyclopedia have been assigned under new metainfo
    within a new section called "encyclopedia". In the future these separate
    metainfos could be absorbed into the existing metainfo hiearchy.
    """
    def calc_type(self, calc: Calculation) -> str:
        """Decides what type of calculation this is: single_point, md,
        geometry_optimization, etc.
        """
        calc_enums = Calculation.calculation_type.type
        calc_type = calc_enums.unavailable

        # Primarily try to determine the calculation type from workflow
        # information
        try:
            workflow = self.entry_archive.section_workflow
            workflow_map = {
                "molecular_dynamics": calc_enums.molecular_dynamics,
                "geometry_optimization": calc_enums.geometry_optimization,
                "phonon": calc_enums.phonon_calculation,
            }
            workflow_enum = workflow_map.get(workflow.workflow_type)
            if workflow_enum is not None:
                calc.calculation_type = workflow_enum
                return workflow_enum
        except Exception:
            pass

        # Fall back to old frame sequence data
        try:
            sccs = self.section_run.section_single_configuration_calculation
        except Exception:
            sccs = []
        try:
            frame_sequences = self.section_run.section_frame_sequence
        except Exception:
            frame_sequences = []

        n_scc = len(sccs)
        n_frame_seq = len(frame_sequences)

        # No sequences, only a few calculations
        if n_scc <= 3 and n_frame_seq == 0:
            program_name = self.section_run.program_name
            if program_name == "elastic":
                # TODO move to taylor expansion as soon as data is correct in archive
                calc_type = calc_enums.elastic_constants
            else:
                calc_type = calc_enums.single_point

        # One sequence. Currently calculations with multiple sequences are
        # unsupported.
        elif n_frame_seq == 1:
            frame_seq = frame_sequences[0]

            # See if sampling_method is present
            section_sampling_method = frame_seq.frame_sequence_to_sampling_ref
            if section_sampling_method is None:
                self.logger.info(
                    "Cannot determine encyclopedia run type because missing "
                    "value for frame_sequence_to_sampling_ref."
                )
                return calc_type

            # See if local frames are present
            frames = frame_seq.frame_sequence_local_frames_ref
            if not frames:
                self.logger.info("No frames referenced in section_frame_sequence_local_frames.")
                return calc_type

            sampling_method = section_sampling_method.sampling_method

            if sampling_method == "molecular_dynamics":
                calc_type = calc_enums.molecular_dynamics
            if sampling_method == "geometry_optimization":
                calc_type = calc_enums.geometry_optimization
            if sampling_method == "taylor_expansion":
                calc_type = calc_enums.phonon_calculation

        calc.calculation_type = calc_type
        return calc_type

    def material_type(self, material: Material) -> tuple:
        # Try to fetch representative system
        system = None
        material_type = config.services.unavailable_value
        material_enums = Material.material_type.type
        try:
            system_idx = self.section_run.m_cache["representative_system_idx"]
        except (AttributeError, KeyError):
            pass
        else:
            # Try to find system type information from archive for the selected system.
            try:
                system = self.section_run.section_system[system_idx]
                stype = system.system_type
            except KeyError:
                pass
            else:
                if stype == material_enums.one_d or stype == material_enums.two_d:
                    material_type = stype
                # For bulk systems we also ensure that the symmetry information is available
                if stype == material_enums.bulk:
                    try:
                        system.section_symmetry[0]
                    except (KeyError, IndexError):
                        self.logger.info("Symmetry information is not available for a bulk system. No Encylopedia entry created.")
                    else:
                        material_type = stype

        material.material_type = material_type
        return system, material_type

    def method_type(self, method: Method) -> tuple:
        repr_method = None
        method_id = config.services.unavailable_value
        methods = self.section_run.section_method
        n_methods = len(methods)

        if n_methods == 1:
            repr_method = methods[0]
            method_id = repr_method.electronic_structure_method
            if method_id is None:
                method_id = config.services.unavailable_value
        elif n_methods > 1:
            for sec_method in methods:
                # GW
                electronic_structure_method = sec_method.electronic_structure_method
                if electronic_structure_method in {"G0W0", "scGW"}:
                    repr_method = sec_method
                    method_id = "GW"
                    break

                # Methods linked to each other through references. Get all
                # linked methods, try to get electronic_structure_method from
                # each.
                try:
                    refs = sec_method.section_method_to_method_refs
                except KeyError:
                    pass
                else:
                    linked_methods = [sec_method]
                    for ref in refs:
                        method_to_method_kind = ref.method_to_method_kind
                        method_to_method_ref = ref.method_to_method_ref
                        if method_to_method_kind == "core_settings":
                            linked_methods.append(method_to_method_ref)

                    for i_method in linked_methods:
                        electronic_structure_method = i_method.electronic_structure_method
                        if electronic_structure_method is not None:
                            repr_method = sec_method
                            method_id = electronic_structure_method

        method.method_type = method_id
        return repr_method, method_id

    def fill(self, context: Context):
        # Fill structure related metainfo
        struct: Any = None
        if context.material_type == Material.material_type.type.bulk:
            struct = MaterialBulkNormalizer(self.entry_archive, self.logger)
        elif context.material_type == Material.material_type.type.two_d:
            struct = Material2DNormalizer(self.entry_archive, self.logger)
        elif context.material_type == Material.material_type.type.one_d:
            struct = Material1DNormalizer(self.entry_archive, self.logger)
        if struct is not None:
            struct.normalize(context)

        # Fill method related metainfo
        method = None
        if context.method_type == Method.method_type.type.DFT or context.method_type == Method.method_type.type.DFTU:
            method = MethodDFTNormalizer(self.entry_archive, self.logger)
        elif context.method_type == Method.method_type.type.GW:
            method = MethodGWNormalizer(self.entry_archive, self.logger)
        if method is not None:
            method.normalize(context)

        # Fill properties related metainfo
        properties = PropertiesNormalizer(self.entry_archive, self.logger)
        properties.normalize(context)

    def normalize(self, logger=None) -> None:
        """The caller will automatically log if the normalizer succeeds or ends
        up with an exception.
        """
        sec_enc = self.entry_archive.section_metadata.m_create(EncyclopediaMetadata)
        status_enums = EncyclopediaMetadata.status.type
        calc_enums = Calculation.calculation_type.type

        # Do nothing if section_run is not present
        if self.section_run is None:
            status = status_enums.invalid_metainfo
            sec_enc.status = status
            self.logger.info(
                "required metainfo is missing or is invalid.",
                enc_status=status,
                invalid_metainfo="section_run",
            )
            return

        try:
            super().normalize(logger)
            # Initialise metainfo structure
            material = sec_enc.m_create(Material)
            method = sec_enc.m_create(Method)
            sec_enc.m_create(Properties)
            calc = sec_enc.m_create(Calculation)

            # Determine run type, stop if unknown
            calc_type = self.calc_type(calc)
            if calc_type == config.services.unavailable_value:
                status = status_enums.unsupported_calculation_type
                sec_enc.status = status
                self.logger.info(
                    "unsupported calculation type for encyclopedia",
                    enc_status=status,
                )
                return

            # Get the system type.
            material_enums = Material.material_type.type
            representative_system, material_type = self.material_type(material)
            if material_type != material_enums.bulk and material_type != material_enums.two_d and material_type != material_enums.one_d:
                status = status_enums.unsupported_material_type
                sec_enc.status = status
                self.logger.info(
                    "unsupported material type for encyclopedia",
                    enc_status=status,
                )
                return

            # Get the method type. For now, we allow unknown method type for
            # phonon calculations, as the method information is resolved at a
            # later stage.
            representative_method, method_type = self.method_type(method)
            if method_type == config.services.unavailable_value:
                sec_enc.status = status_enums.unsupported_method_type
                self.logger.info(
                    "unsupported method type for encyclopedia",
                    enc_status=status_enums.unsupported_method_type,
                )
                if calc_type != calc_enums.phonon_calculation:
                    return

            # Get representative scc
            try:
                representative_scc_idx = self.section_run.m_cache["representative_scc_idx"]
                representative_scc = self.section_run.section_single_configuration_calculation[representative_scc_idx]
            except Exception:
                representative_scc = None
                representative_scc_idx = None

            # Create one context that holds all details
            context = Context(
                material_type=material_type,
                method_type=method_type,
                calc_type=calc_type,
                representative_system=representative_system,
                representative_method=representative_method,
                representative_scc=representative_scc,
                representative_scc_idx=representative_scc_idx,
            )

            # Put the encyclopedia section into archive
            self.fill(context)

            # Check that the necessary information is in place
            functional_type = method.functional_type
            if functional_type is None:
                sec_enc.status = status_enums.unsupported_method_type
                self.logger.info(
                    "unsupported functional type for encyclopedia",
                    enc_status=status_enums.unsupported_method_type,
                )
                return

        except Exception:
            status = status_enums.failure
            sec_enc.status = status
            self.logger.error(
                "failed to process encyclopedia data due to an unhandlable exception",
                enc_status=status,
            )
            raise  # Reraise for the caller to log the exception as well
        else:
            status = status_enums.success
            sec_enc.status = status
            self.logger.info(
                "successfully created metainfo for encyclopedia.",
                enc_status=status,
            )



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class CPMDParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*\-*\+*\d*'
        super().__init__(
            specifications=dict(
                name='parsers/cpmd', code_name='CPMD',
                code_homepage='https://www.lcrc.anl.gov/for-users/software/available-software/cpmd/',
                mainfile_contents_re=(r'\*\*\*       \*\*   \*\*\*  \*\* \*\*\*\* \*\*  \*\*   \*\*\*')),
            units_mapping=dict(length=ureg.bohr, energy=ureg.hartree),
            program_version=r'\*\s*VERSION\s*(\d+\.\d+.+)',
            atom_labels_atom_positions_atom_forces=r'ATOM\s*COORDINATES\s*GRADIENTS([\s\S]+?)(?:\*{50}|\n *\n)',
            atom_labels_atom_positions=r'ATOMIC COORDINATES\s*\*\s*\*+([\s\S]+?)\*{50}',
            lattice_vectors=(
                rf'(LATTICE VECTOR A1\(BOHR\):\s*{re_f}\s+{re_f}\s+{re_f}\s+)'
                rf'(LATTICE VECTOR A2\(BOHR\):\s*{re_f}\s+{re_f}\s+{re_f}\s+)'
                rf'(LATTICE VECTOR A3\(BOHR\):\s*{re_f}\s+{re_f}\s+{re_f}\s+)'),
            energy_total=rf' ETOT\s*\=\s*({re_f})|\(K\+E1\+L\+N\+X\)\s*TOTAL ENERGY\s*\=\s*({re_f})',
            electronic_kinetic_energy=rf'\(K\)\s*KINETIC ENERGY\s*=\s*({re_f})',
            energy_electrostatic=rf'\(E1\=A\-S\+R\)\s*ELECTROSTATIC ENERGY\s*=\s*({re_f})',
            energy_XC=rf'\(X\)\s*EXCHANGE\-CORRELATION ENERGY\s*=\s*({re_f})')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from rdflib import Graph, Literal, RDF, URIRef, BNode
from rdflib.namespace import Namespace, DCAT, DCTERMS as DCT, FOAF, RDF

from nomad import config
from nomad.datamodel import User

from nomad.datamodel import EntryMetadata, User

from .api import url

VCARD = Namespace('http://www.w3.org/2006/vcard/ns#')
HYDRA = Namespace('http://www.w3.org/ns/hydra/core#')


def get_optional_entry_prop(entry, name):
    try:
        return entry[name]
    except (KeyError, AttributeError):
        return 'unavailable'


class Mapping():
    def __init__(self):
        self.g = Graph()
        self.g.bind('rdf', RDF)
        self.g.bind('dcat', DCAT)
        self.g.bind('dct', DCT)
        self.g.bind('vcard', VCARD)
        self.g.bind('foaf', FOAF)
        self.g.bind('hydra', HYDRA)

        self.persons = {}

    def map_catalog(self, entries, after: str, modified_since, slim=True):
        def uri_ref(after):
            kwargs = dict()
            if after is not None:
                kwargs['after'] = after
            if modified_since is not None:
                kwargs['modified_since'] = modified_since.strftime('%Y-%m-%d')
            return URIRef(url('catalog', **kwargs))

        after = after.strip()

        catalog = uri_ref(after=None)
        self.g.add((catalog, RDF.type, DCAT.Catalog))
        last_entry = None
        for entry in entries:
            self.g.add((catalog, DCT.dataset, self.map_entry(entry, slim=slim)))
            last_entry = entry

        hydra_collection = uri_ref(after)
        self.g.add((hydra_collection, RDF.type, HYDRA.Collection))
        self.g.add((hydra_collection, HYDRA.totalItems, Literal(entries.total)))
        self.g.add((hydra_collection, HYDRA.first, uri_ref('')))
        if last_entry is not None:
            self.g.add((hydra_collection, HYDRA.next, uri_ref(last_entry.calc_id)))

        self.g.add((hydra_collection, RDF.type, HYDRA.collection))

        for person in self.persons.values():
            self.g.add((catalog, DCT.creator, person))

    def map_entry(self, entry: EntryMetadata, slim=False):
        dataset = URIRef(url('datasets', entry.calc_id))

        self.g.add((dataset, RDF.type, DCAT.Dataset))
        self.g.add((dataset, DCT.identifier, Literal(entry.calc_id)))
        self.g.add((dataset, DCT.issued, Literal(entry.upload_time)))
        self.g.add((dataset, DCT.modified, Literal(entry.last_processing)))
        self.g.add((dataset, DCT.title, Literal(get_optional_entry_prop(entry, 'formula'))))
        self.g.add((dataset, DCT.description, Literal(get_optional_entry_prop(entry, 'comment'))))

        if slim:
            return dataset

        self.g.add((dataset, DCAT.landingPage, URIRef('%s/entry/id/%s/%s' % (
            config.gui_url(), entry.upload_id, entry.calc_id))))

        self.g.add((dataset, DCT.license, URIRef('https://creativecommons.org/licenses/by/4.0/legalcode')))
        self.g.add((dataset, DCT.language, URIRef('http://id.loc.gov/vocabulary/iso639-1/en')))

        self.g.add((dataset, DCT.publisher, self.map_user(entry.uploader)))
        try:
            for author in entry.authors:
                self.g.add((dataset, DCT.creator, self.map_user(author)))
        except (KeyError, AttributeError):
            pass
        self.g.add((dataset, DCAT.contactPoint, self.map_contact(entry.uploader)))

        self.g.add((dataset, DCAT.distribution, self.map_distribution(entry, 'api')))
        self.g.add((dataset, DCAT.distribution, self.map_distribution(entry, 'json')))
        self.g.add((dataset, DCAT.distribution, self.map_distribution(entry, 'raw')))

        return dataset

    def map_user(self, user: User):
        person = self.persons.get(user.user_id)
        if person is not None:
            return person

        user = User.get(user.user_id)
        person = BNode()

        self.g.add((person, RDF.type, FOAF.Person))
        self.g.add((person, FOAF.givenName, Literal(user.first_name)))
        self.g.add((person, FOAF.familyName, Literal(user.last_name)))
        self.g.add((person, FOAF.nick, Literal(user.username)))
        self.g.add((person, FOAF.mbox, URIRef('mailto:%s' % (user.email))))

        self.persons[user.user_id] = person

        return person

    def map_contact(self, user: User):
        person = self.persons.get(user.user_id)
        if person is None:
            person = self.map_user(user)

        user = User.get(user.user_id)
        self.g.add((person, RDF.type, VCARD.Individual))
        self.g.add((person, VCARD.givenName, Literal(user.first_name)))
        self.g.add((person, VCARD.familyName, Literal(user.last_name)))
        self.g.add((person, VCARD.nickName, Literal(user.username)))
        self.g.add((person, VCARD.hasEmail, Literal(user.email)))
        self.g.add((person, VCARD.organization, Literal(get_optional_entry_prop(user, 'affiliation'))))
        # address = BNode()
        # self.g.add((address, RDF.type, VCARD.Address))
        # self.g.add((address, VCARD.street_address, )) # affiliation_address?
        # self.g.add((address, VCARD.postal_code, )) # affiliation_address?
        # self.g.add((address, VCARD.country_name, )) # affiliation_address?
        # self.g.add((address, VCARD.locality, )) # affiliation_address?
        # self.g.add((address, VCARD.region, )) # affiliation_address?
        # self.g.add((person, VCARD.hasAddress, address))

        return person

    def map_distribution(self, entry, dist_kind):
        if dist_kind == 'api':
            # DataService: API
            service = BNode()
            self.g.add((service, RDF.type, DCAT.DataService))
            self.g.add((service, DCT.title, Literal('NOMAD API')))  # How to include terms from swagger document here?
            self.g.add((service, DCT.description, Literal('Official NOMAD API')))  # same question
            self.g.add((service, DCAT.endpointURL, URIRef('https://nomad-lab.eu/prod/rae/api/')))  # config.api_url() ?
            # not sure if the following needs to be dataset specific:
            self.g.add((service, DCAT.endpointDescription, URIRef('https://nomad-lab.eu/prod/rae/api/swagger.json')))

            # Distribution over API
            dist = BNode()
            self.g.add((dist, DCT.title, Literal(get_optional_entry_prop(entry, 'formula') + '_api')))
            self.g.add((dist, RDF.type, DCAT.Distribution))
            self.g.add((dist, DCAT.accessService, service))
        elif dist_kind == 'json':
            # Distribution as JSON
            dist = BNode()
            self.g.add((dist, RDF.type, DCAT.Distribution))
            self.g.add((dist, DCT.title, Literal(get_optional_entry_prop(entry, 'formula') + '_json')))
            self.g.add((dist, DCAT.mediaType, URIRef('https://www.iana.org/assignments/media-types/application/json')))
            self.g.add((dist, DCAT.packageFormat, URIRef('https://www.iana.org/assignments/media-types/application/zip')))
            self.g.add((dist, DCAT.downloadURL, URIRef(
                'http://nomad-lab.eu/prod/rae/api/archive/download?upload_id=%s&calc_id=%s' % (entry.upload_id, entry.calc_id))))
            self.g.add((dist, DCAT.accessURL, URIRef('%s/entry/id/%s/%s' % (
                config.gui_url(), entry.upload_id, entry.calc_id))))
        elif dist_kind == 'raw':
            # Distribution of the raw data
            dist = BNode()
            self.g.add((dist, RDF.type, DCAT.Distribution))
            self.g.add((dist, DCT.title, Literal(get_optional_entry_prop(entry, 'formula') + '_raw')))
            self.g.add((dist, DCAT.accessURL, URIRef('https://nomad-lab.eu/prod/rae/api/raw/calc/%s/%s' % (
                entry.upload_id, entry.calc_id))))
            self.g.add((dist, DCAT.packageFormat, URIRef('https://www.iana.org/assignments/media-types/application/zip')))

        return dist



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class OnetepParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*e*\-*\+*\d*'

        super().__init__(
            specifications=dict(
                name='parsers/onetep', code_name='ONETEP', code_homepage='https://www.onetep.org/',
                domain='dft', mainfile_contents_re=r'####### #     # ####### ####### ####### ######'),
            units_mapping=dict(energy=ureg.hartree, length=ureg.bohr),
            auxilliary_files=r'([\w\-]+\.dat)',
            program_version=r'Version\s*([\d\.]+)',
            lattice_vectors=r'\%block lattice_cart\s*([\s\S]+?)\%endblock lattice_cart',
            atom_labels_atom_positions=rf'\%block positions\_abs\s*(\w+\s+{re_f}\s+{re_f}\s+{re_f}[\s\S]+?)\%endblock positions\_abs',
            XC_functional=r'xc\_functional\s*\:\s*(\w+)',
            energy_total=rf'Total energy\s*=\s*({re_f})\s*Eh',
            atom_forces=r'Forces \*+([\s\S]+?)\*{50}')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click
import os
import shutil
import tabulate
import elasticsearch_dsl

from nomad import config as nomad_config, infrastructure, processing
from nomad import search as nomad_search

from .admin import admin


@admin.command(help='Checks consistency of files and es vs mongo and deletes orphan entries.')
@click.option('--dry', is_flag=True, help='Do not delete anything, just check.')
@click.option('--skip-calcs', is_flag=True, help='Skip cleaning calcs with missing uploads.')
@click.option('--skip-fs', is_flag=True, help='Skip cleaning the filesystem.')
@click.option('--skip-es', is_flag=True, help='Skip cleaning the es index.')
@click.option('--staging-too', is_flag=True, help='Also clean published entries in staging, make sure these files are not due to reprocessing')
@click.option('--force', is_flag=True, help='Do not ask for confirmation.')
def clean(dry, skip_calcs, skip_fs, skip_es, staging_too, force):
    mongo_client = infrastructure.setup_mongo()
    infrastructure.setup_elastic()

    if not skip_calcs:
        uploads_for_calcs = mongo_client[nomad_config.mongo.db_name]['calc'].distinct('upload_id')
        uploads = {}
        for upload in mongo_client[nomad_config.mongo.db_name]['upload'].distinct('_id'):
            uploads[upload] = True

        missing_uploads = []
        for upload_for_calc in uploads_for_calcs:
            if upload_for_calc not in uploads:
                missing_uploads.append(upload_for_calc)

        if not dry and len(missing_uploads) > 0:
            if not force:
                input('Will delete calcs (mongo + es) for %d missing uploads. Press any key to continue ...' % len(missing_uploads))

            for upload in missing_uploads:
                mongo_client[nomad_config.mongo.db_name]['calc'].remove(dict(upload_id=upload))
                elasticsearch_dsl.Search(index=nomad_config.elastic.index_name).query('term', upload_id=upload).delete()
        else:
            print('Found %s uploads that have calcs in mongo, but there is no upload entry.' % len(missing_uploads))
            print('List first 10:')
            for upload in missing_uploads[:10]:
                print(upload)

    if not skip_fs:
        upload_dirs = []
        public_dirs, staging_dirs = {}, {}
        for bucket, dir_map in [(nomad_config.fs.public, public_dirs), (nomad_config.fs.staging, staging_dirs)]:
            for prefix in os.listdir(bucket):
                for upload in os.listdir(os.path.join(bucket, prefix)):
                    upload_dirs.append((upload, os.path.join(bucket, prefix, upload)))
                    dir_map[upload] = os.path.join(bucket, prefix, upload)

        to_delete = list(
            path for upload, path in upload_dirs
            if processing.Upload.objects(upload_id=upload).first() is None)

        if not dry and len(to_delete) > 0:
            if not force:
                input('Will delete %d upload directories. Press any key to continue ...' % len(to_delete))

            for path in to_delete:
                shutil.rmtree(path)
        else:
            print('Found %d upload directories with no upload in mongo.' % len(to_delete))
            print('List first 10:')
            for path in to_delete[:10]:
                print(path)

    if staging_too and not skip_fs:
        to_delete = list(
            path for upload, path in staging_dirs.items()
            if upload in public_dirs)

        if not dry and len(to_delete) > 0:
            if not force:
                input('Will delete %d staging upload directories. Press any key to continue ...' % len(to_delete))

            for path in to_delete:
                shutil.rmtree(path)
        else:
            print('Found %d staging upload directories with upload directory in public.' % len(to_delete))
            print('List first 10:')
            for path in to_delete[:10]:
                print(path)

    if not skip_es:
        search = nomad_search.Search(index=nomad_config.elastic.index_name)
        search.aggs.bucket('uploads', elasticsearch_dsl.A('terms', field='upload_id', size=12000))
        response = search.execute()

        to_delete = list(
            (bucket.key, bucket.doc_count)
            for bucket in response.aggregations.uploads.buckets
            if processing.Upload.objects(upload_id=bucket.key).first() is None)

        calcs = 0
        for _, upload_calcs in to_delete:
            calcs += upload_calcs

        if not dry and len(to_delete) > 0:
            if not force:
                input(
                    'Will delete %d calcs in %d uploads from ES. Press any key to continue ...' %
                    (calcs, len(to_delete)))
            for upload, _ in to_delete:
                nomad_search.Search(index=nomad_config.elastic.index_name).query('term', upload_id=upload).delete()
        else:
            print('Found %d calcs in %d uploads from ES with no upload in mongo.' % (calcs, len(to_delete)))
            print('List first 10:')
            tabulate.tabulate(to_delete, headers=['id', '#calcs'])



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import logging
import pint
from typing import Any, Dict
import gzip
import bz2
import lzma
import tarfile


class FileParser:
    '''
    Base class for parsers. The parse method specific to a file type
    should be implemented in the corresponding child class. The parsed quantities are
    stored in results. One can access a quantity by using the get method or as attribute.

    Arguments:
        mainfile: the file to be parsed
        logger: optional logger
    '''
    def __init__(self, mainfile=None, logger=None, open=None):
        self._mainfile: Any = None
        self._mainfile_obj: Any = None
        if isinstance(mainfile, str):
            self._mainfile = os.path.abspath(mainfile)
            self._mainfile_obj = None
        elif hasattr(mainfile, 'name'):
            self._mainfile = mainfile.name
            self._mainfile_obj = mainfile
        self._open = open
        self.logger = logger if logger is not None else logging
        self._results: Dict[str, Any] = None
        # a key is necessary for xml parsers, where parsing is done dynamically
        self._key: str = None
        self._kwargs: Dict[str, Any] = dict()
        self._file_handler: Any = None

    def init_parameters(self):
        pass

    @property
    def results(self):
        if self._results is None:
            self._results = dict()
        if self._key not in self._results:
            self.parse(self._key, **self._kwargs)

        return self._results

    @property
    def maindir(self):
        return os.path.dirname(self._mainfile)

    @property
    def mainfile_obj(self):
        if self._mainfile_obj is None:
            try:
                self._mainfile_obj = self.open(self._mainfile)
            except Exception:
                pass

        return self._mainfile_obj

    @property
    def mainfile(self):
        if self._mainfile is None:
            return

        if self._mainfile_obj is None and not os.path.isfile(self._mainfile):
            return
        return self._mainfile

    @mainfile.setter
    def mainfile(self, val):
        self._results = None
        self._file_handler = None
        self._mainfile = None
        if isinstance(val, str):
            self._mainfile = os.path.abspath(val)
            self._mainfile_obj = None
        elif hasattr(val, 'name'):
            self._mainfile = val.name
            self._mainfile_obj = val
        self.init_parameters()

    def open(self, mainfile):
        open_file = self._open
        if open_file is None:
            if mainfile.endswith('.gz'):
                open_file = gzip.open
            elif mainfile.endswith('.bz2'):
                open_file = bz2.open
            elif mainfile.endswith('.xz'):
                open_file = lzma.open
            elif mainfile.endswith('.tar'):
                open_file = tarfile.open
            else:
                open_file = open
        return open_file(mainfile)

    def get(self, key: str, default: Any = None, unit: Any = None, **kwargs):
        '''
        Returns the parsed result for quantity with name key. If quantity is not in
        results default will be returned. A pint unit can be provided which is attached
        to the returned value.
        '''
        if self.mainfile is None:
            return default

        self._key = key
        self._kwargs = kwargs
        val = self.results.get(key, None)
        if val is None:
            val = default

        if val is None:
            return

        if unit is not None:
            if isinstance(unit, pint.Quantity) or isinstance(unit, pint.Unit):
                val = val * unit

            elif isinstance(val, pint.Quantity):
                val = val.to(unit)

            else:
                val = pint.Quantity(val, unit)

        return val

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.get(key)
        elif isinstance(key, int):
            return self[int]

    def __getattr__(self, key):
        if self._results is None:
            self._results = dict()

        return self._results.get(key, None)

    def parse(self, quantity_key: str = None, **kwargs):
        pass



import re
import numpy as np
import logging
import ase

from .metainfo import m_env

from nomad.units import ureg
from nomad.parsing import FairdiParser
from nomad.parsing.file_parser.text_parser import TextParser, Quantity
from nomad.datamodel.metainfo.common_dft import Run, Method, System, XCFunctionals,\
    BasisSetAtomCentered, SingleConfigurationCalculation, Eigenvalues, SamplingMethod,\
    ScfIteration
from .metainfo.gaussian import x_gaussian_section_elstruc_method,\
    x_gaussian_section_moller_plesset, x_gaussian_section_hybrid_coeffs,\
    x_gaussian_section_coupled_cluster, x_gaussian_section_quadratic_ci,\
    x_gaussian_section_ci, x_gaussian_section_semiempirical, x_gaussian_section_molmech,\
    x_gaussian_section_excited, x_gaussian_section_casscf, x_gaussian_section_orbital_symmetries,\
    x_gaussian_section_geometry_optimization_info, x_gaussian_section_molecular_multipoles,\
    x_gaussian_section_frequencies, x_gaussian_section_thermochem,\
    x_gaussian_section_force_constant_matrix, x_gaussian_section_times,\
    x_gaussian_section_symmetry


class GaussianOutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_float = r'[\d\.\-\+Ee]+'
        re_float_dexp = r'[\d\.\-\+EeDd]+'
        re_eigs = re.compile(r'(\-?\d*\.\d+\s*)')
        re_force = re.compile(r'(\d+\s*\-?\d*\..*)')

        def str_to_exp(val_in):
            val = np.array(val_in.rstrip('.').upper().replace('D', 'E').split(), dtype=float)
            return val[0] if len(val) == 1 else val

        def str_to_orbital_symmetries(val_in):
            val = re.findall(r'(?:Occupied|Virtual)\s*((?:\(.+\)\s*)*)', val_in)
            return [v.replace('(', '').replace(')', '').split() for v in val]

        def str_to_eigenvalues(val_in):
            val = val_in.split()
            spin_index = 0 if val[0] == 'Alpha' else 1
            occ = 0. if val[1] == 'virt' else 1.
            eigs, occs = [[], []], [[], []]
            # gaussian sometimes prints values without spaces so we need a re matching
            eigs[spin_index] = re_eigs.findall(val_in)
            occs[spin_index] = [occ] * len(eigs[spin_index])
            return eigs, occs

        def str_to_normal_modes(val_in):
            val = [v.split() for v in val_in.split('\n')]
            return np.array([v[2:] for v in val if len(v) >= 5], dtype=float)

        def str_to_force_constants(val_in):
            val = re_force.findall(val_in)
            fc = []
            for v in val:
                v = np.array(v.upper().replace('D', 'E').split(), dtype=float)
                index = int(v[0])
                if len(fc) < index:
                    fc.append([])
                fc[index - 1].extend(v[1:])
            for n in range(len(fc)):
                fc[n].extend(np.zeros(index - n - 1))
            fc = np.array(fc)
            fc = fc + fc.T - np.diag(fc.diagonal())
            return fc

        orientation_quantities = [
            Quantity(
                'standard_orientation',
                r'd orientation[\s\S]+?X\s*Y\s*Z\s*\-+\s*([\d\.\s\-]+?)\-{2}',
                convert=False, str_operation=lambda x: np.array(
                    [v.split() for v in x.strip().split('\n')], dtype=float)),
            Quantity(
                'input_orientation',
                r't orientation[\s\S]+?X\s*Y\s*Z\s*\-+\s*([\d\.\s\-]+?)\-{2}',
                convert=False, str_operation=lambda x: np.array(
                    [v.split() for v in x.strip().split('\n')], dtype=float)),
            Quantity(
                'z_matrix_orientation',
                r'x orientation[\s\S]+?X\s*Y\s*Z\s*\-+\s*([\d\.\s\-]+?)\-{2}',
                convert=False, str_operation=lambda x: np.array(
                    [v.split() for v in x.strip().split('\n')], dtype=float))]

        calculation_quantities = [
            Quantity(
                'energy_total',
                rf'\n *(?:Energy=|Electronic Energy)\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'hybrid_xc_coeff1', rf'ScaHFX=\s*({re_float})', dtype=float),
            Quantity(
                'hybrid_xc_coeff2',
                rf'ScaDFX=\s*({re_float}\s*{re_float}\s*{re_float}\s*{re_float})',
                flatten=False),
            Quantity('mp', r'(E2) ='),
            Quantity(
                'energy_total_mp',
                rf'(?:EUMP2|EUMP3|UMP4\(DQ\)|UMP4\(SDQ\)|UMP4\(SDTQ\)|MP5)\s*=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'mp2_correction_energy',
                rf'E2 =\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'mp3_correction_energy',
                rf'E3\s*=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'mp4dq_correction_energy',
                rf'E4\(DQ\)=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'mp4sdq_correction_energy',
                rf'E4\(SDQ\)=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'mp4sdtq_correction_energy',
                rf'E4\(SDTQ\)=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'mp5_correction_energy',
                rf'DEMP5 =\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity('cc', r'((?:CCSD\(T\)|E\(CORR\)))'),
            Quantity(
                'energy_total_cc',
                rf'(?:\n *CCSD\(T\)|E\(CORR\))\s*=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'ccsd_correction_energy',
                rf'DE\(Corr\)=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity('qci', r'(Quadratic Configuration Interaction)'),
            Quantity(
                'energy_total_qci',
                rf'(?:QCISD\(T\)|E\(Z\)|QCISD\(TQ\))=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'qcisd_correction_energy',
                rf'DE\(Z\)=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'qcisdtq_correction_energy',
                rf'DE5\s*=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity('ci', r'(\s{2}Configuration Interaction)'),
            Quantity(
                'energy_total_ci',
                rf'E\(CI\)=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'ci_correction_energy',
                rf'DE\(CI\)=\s*({re_float_dexp})',
                str_operation=str_to_exp, convert=False, unit='hartree', repeats=True),
            Quantity(
                'semiempirical_method',
                r'([-A-Z0-9]+\s*calculation of energy[a-zA-Z,. ]+)'),
            Quantity(
                'semiempirical_energy',
                rf'It=\s*\d+\s*PL=\s*[-+0-9EeDd.]+\s*DiagD=[A-Z]\s*ESCF=\s*({re_float})',
                repeats=True, unit='hartree', dtype=float),
            Quantity(
                'molmech_method',
                r'([a-zA-Z0-9]+\s*calculation of energy[a-z,. ]+)'),
            Quantity(
                'excited_state',
                rf'Excited State\s*(\d+):\s*\S+\s*({re_float})\s*eV\s*{re_float}\s*nm\s*'
                rf'f=({re_float})\s*<[\w*]+>=({re_float})\s*(.*)',
                repeats=True, convert=False),
            Quantity(
                'casscf_energy',
                rf'\(\s*[0-9]+\)\s*EIGENVALUE\s*({re_float})',
                repeats=True, dtype=float, unit='hartree'),
            Quantity(
                'optimization_completed',
                r'(Optimization (?:completed|stopped))', flatten=False, convert=False),
            Quantity(
                'population_analysis',
                r'(Population analysis using the SCF density[\s\S]+?Condensed)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'orbital_symmetries',
                        r'Orbital symmetries:([\s\S]*?)electro',
                        repeats=False, str_operation=str_to_orbital_symmetries),
                    Quantity(
                        'x_gaussian_elstate_symmetry',
                        r'nic state is\s*(.+)\.', flatten=False),
                    Quantity(
                        'eigenvalues',
                        r'(Alpha|Beta)\s*(occ|virt)\. eigenvalues \-\-\s*(.+)',
                        repeats=True, str_operation=str_to_eigenvalues, convert=False)])),
            Quantity(
                'charge',
                rf'\n *Charge=\s*({re_float})\s*electrons', dtype=float, unit='elementary_charge'),
            Quantity(
                'dipole',
                r''.join([r'%s=\s*([\-\d\.]+)\s*' % c for c in ['X', 'Y', 'Z']]),
                dtype=float, unit='debye'),
            Quantity(
                'quadrupole',
                r''.join([r'%s=\s*([\-\d\.]+)\s*' % c for c in [
                    'XX', 'YY', 'ZZ', 'XY', 'XZ', 'YZ']]),
                dtype=float, unit='debye * angstrom'),
            Quantity(
                'octapole',
                r''.join([r'%s=\s*([\-\d\.]+)\s*' % c for c in [
                    'XXX', 'YYY', 'ZZZ', 'XYY', 'XXY', 'XXZ', 'XZZ', 'YZZ', 'YYZ', 'XYZ']]),
                dtype=float, unit='debye * angstrom**2'),
            Quantity(
                'hexadecapole',
                r''.join([r'%s=\s*([\-\d\.]+)\s*' % c for c in [
                    'XXXX', 'YYYY', 'ZZZZ', 'XXXY', 'XXXZ', 'YYYX', 'YYYZ', 'ZZZX',
                    'ZZZY', 'XXYY', 'XXZZ', 'YYZZ', 'XXYZ', 'YYXZ', 'ZZXY']]),
                dtype=float, unit='debye * angstrom**3'),
            Quantity(
                'frequencies',
                r'\n *Frequencies \-\-\s*(.+)', dtype=float, repeats=True),
            Quantity(
                'reduced_masses',
                r'\n *Red\. masses \-\-\s*(.+)', dtype=float, repeats=True),
            Quantity(
                'normal_modes',
                r'Atom\s*AN.*\s*([\-\d\s\.]+)',
                str_operation=str_to_normal_modes, convert=False, repeats=True),
            Quantity(
                'temperature_pressure',
                rf'Temperature\s*({re_float})\s*Kelvin\.\s*Pressure\s*({re_float})\s*Atm\.'),
            Quantity(
                'moments',
                r'(?:Eigenvalues|EIGENVALUES) \-\-\s*(\d+\.\d{5})\s*(\d+\.\d{5})\s*(\d+\.\d{5})',
                dtype=float, unit='amu*angstrom**2'),
            Quantity(
                'zero_point_energy',
                rf'Zero\-point correction=\s*({re_float})',
                dtype=float, unit='hartree'),
            Quantity(
                'thermal_correction_energy',
                rf'Thermal correction to Energy=\s*({re_float})',
                dtype=float, unit='hartree'),
            Quantity(
                'thermal_correction_enthalpy',
                rf'Thermal correction to Enthalpy=\s*({re_float})',
                dtype=float, unit='hartree'),
            Quantity(
                'thermal_correction_free_energy',
                rf'Thermal correction to Gibbs Free Energy=\s*({re_float})',
                dtype=float, unit='hartree'),
            Quantity(
                'forces',
                r'Forces \(Hartrees/Bohr\)\s*Number\s*Number\s*X\s*Y\s*Z\s*\-+\s*([\d\s\-\.]+?)\s*\-\-',
                str_operation=lambda x: np.array([xi.split()[2:5] for xi in x.split('\n')], dtype=float),
                convert=False, unit='hartree/bohr'),
            Quantity(
                'force_constants',
                r'Force constants in Cartesian coordinates:\s*([\s\S]+?)Force',
                str_operation=str_to_force_constants, convert=False),
            Quantity(
                'scf_iteration',
                r'(cle\s*\d+[\s\S]+?(?:Cy|\n\n|Leave))', sub_parser=TextParser(quantities=[
                    Quantity(
                        'number', r'cle\s*(\d+)', dtype=int),
                    Quantity(
                        'energy_total_scf_iteration',
                        rf' E=\s*({re_float})', dtype=float, unit='hartree'),
                    Quantity(
                        'x_gaussian_delta_energy_total_scf_iteration',
                        rf'Delta\-E=\s*({re_float})', dtype=float, unit='hartree')]),
                repeats=True),
            Quantity(
                'scf_iteration_final',
                r'(SCF Done:[\s\S]+?(?:Leave|\n\n|End))', sub_parser=TextParser(quantities=[
                    Quantity(
                        'x_gaussian_single_configuration_calculation_converged',
                        r'(SCF Done)', convert=False, flatten=False),
                    Quantity(
                        'x_gaussian_hf_detect', r'E\((.+?)\)'),
                    Quantity(
                        'x_gaussian_energy_scf',
                        rf'({re_float})\s*A\.U\.', dtype=float, unit='hartree'),
                    Quantity(
                        'x_gaussian_energy_error',
                        rf'Conv\s*=\s*({re_float_dexp})', unit='hartree', convert=False,
                        str_operation=str_to_exp),
                    Quantity(
                        'x_gaussian_electronic_kinetic_energy',
                        rf'KE\s*=\s*({re_float_dexp})', unit='hartree', convert=False,
                        str_operation=str_to_exp),
                    Quantity(
                        'x_gaussian_spin_S2', rf'before annihilation\s*({re_float})', dtype=float),
                    Quantity(
                        'x_gaussian_after_annihilation_spin_S2',
                        rf',\s*after\s*({re_float})', dtype=float),
                    Quantity(
                        'x_gaussian_perturbation_energy',
                        r'[()A-Z0-9]+\s*=\s*[-+0-9D.]+\s*[()A-Z0-9]+\s*=\s*([-+0-9D.]+)',
                        convert=False, str_operation=str_to_exp, unit='hartree')]), repeats=False)
        ]

        run_quantities = [
            Quantity(
                'x_gaussian_settings_corrected',
                r'\-{10}\s*(#[\s\S]+?)\-{10}',
                convert=False,
                str_operation=lambda x: x.strip().replace('\n', '')),
            Quantity(
                'x_gaussian_total_charge', r'Charge =\s*([\-\+\d]+)', dtype=int),
            Quantity(
                'x_gaussian_spin_target_multiplicity', r'Multiplicity =\s*([\-\+\d]+)', dtype=int),
            Quantity(
                'lattice_vector', r'(?:TV|Tv)\s*0?\s*([\d\. ]+)', dtype=float, repeats=True),
            Quantity(
                'x_gaussian_atomic_masses', r'IAtWgt=([ \d\.]+)', dtype=float, repeats=True),
            Quantity(
                'system',
                r'((?:Standard|Z-Matrix|Input) orientation:[\s\S]+?)'
                r'(?:Predicted change in Energy|PREDICTED CHANGE IN ENERGY|Z-Matrix orientation|Normal termination)',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'calculation',
                        r'(<S\*\*2> of initial guess=[\s\S]+?(?:Initial guess read from the read\-write|\Z))',
                        sub_parser=TextParser(quantities=calculation_quantities), repeats=True)
                ] + calculation_quantities + orientation_quantities)),
            Quantity(
                'iteration',
                r'ation Nr\.([\s\S]+?)(?:Iter|\n\n)',
                repeats=True, sub_parser=TextParser(quantities=calculation_quantities)),
            Quantity('program_cpu_time', r'Job cpu time:(.*)\.\n', flatten=False),
            Quantity(
                'program_termination_date',
                r'Normal termination of Gaussian\s*\d+\s*at\s*(.*)\.\n', flatten=False)
        ]

        self._quantities = [
            Quantity(
                'program',
                r'\s*Gaussian\s*([0-9]+):\s*(\S+)\s*(\S+)\s*(\d+\-\w+\-\d+)',
                convert=False),
            Quantity('x_gaussian_chk_file', r'%[Cc]hk=([A-Za-z0-9.]*)', dtype=str),
            Quantity('x_gaussian_memory', r'%[Mm]em=([A-Za-z0-9.]*)', dtype=str),
            Quantity(
                'x_gaussian_number_of_processors',
                r'%[Nn][Pp]roc=([A-Za-z0-9.]*)', dtype=str),
            Quantity(
                'run',
                r'(-{10}\s*#[\s\S]+?Normal termination.*\n)', repeats=True,
                sub_parser=TextParser(quantities=run_quantities)),
        ]


class GaussianParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/gaussian', code_name='Gaussian', code_homepage='http://gaussian.com/',
            mainfile_mime_re=r'.*', mainfile_contents_re=(
                r'\s*Cite this work as:'
                r'\s*Gaussian [0-9]+, Revision [A-Za-z0-9\.]*,'))

        self._metainfo_env = m_env
        self.out_parser = GaussianOutParser()

        self._xc_functional_map = {
            'S': [{'name': 'LDA_X'}],
            'XA': [{'name': 'LDA_X_EMPIRICAL'}],
            'VWN': [{'name': 'LDA_C_VWN'}],
            'VWN3': [{'name': 'LDA_C_VWN_3'}],
            'SVWN': [{'name': 'LDA_X'}, {'name': 'LDA_C_VWN'}],
            'LSDA': [{'name': 'LDA_X'}, {'name': 'LDA_C_VWN'}],
            'B': [{'name': 'GGA_X_B88'}],
            'BLYP': [{'name': 'GGA_C_LYP'}, {'name': 'GGA_X_B88'}],
            'PBEPBE': [{'name': 'GGA_C_PBE'}, {'name': 'GGA_X_PBE'}],
            'PBEH': [{'name': 'GGA_X_WPBEH'}],
            'WPBEH': [{'name': 'GGA_X_WPBEH'}],
            'PW91PW91': [{'name': 'GGA_C_PW91'}, {'name': 'GGA_X_PW91'}],
            'M06L': [{'name': 'MGGA_C_M06_L'}, {'name': 'MGGA_X_M06_L'}],
            'M11L': [{'name': 'MGGA_C_M11_L'}, {'name': 'MGGA_X_M11_L'}],
            'SOGGA11': [{'name': 'GGA_X_SOGGA11'}, {'name': 'GGA_C_SOGGA11'}],
            'MN12L': [{'name': 'MGGA_X_MN12_L'}, {'name': 'MGGA_C_MN12_L'}],
            'N12': [{'name': 'GGA_C_N12'}, {'name': 'GGA_X_N12'}],
            'VSXC': [{'name': 'MGGA_X_GVT4'}, {'name': 'MGGA_C_VSXC'}],
            'HCTH93': [{'name': 'GGA_XC_HCTH_93'}],
            'HCTH147': [{'name': 'GGA_XC_HCTH_147'}],
            'HCTH407': [{'name': 'GGA_XC_HCTH_407'}],
            'HCTH': [{'name': 'GGA_XC_HCTH_407'}],
            'B97D': [{'name': 'GGA_XC_B97_D'}],
            'B97D3': [{'name': 'GGA_XC_B97_D3'}],
            'MPW': [{'name': 'GGA_X_MPW91'}],
            'G96': [{'name': 'GGA_X_G96'}],
            'O': [{'name': 'GGA_X_OPTX'}],
            'BRX': [{'name': 'MGGA_X_BR89'}],
            'PKZB': [{'name': 'MGGA_C_PKZB'}, {'name': 'MGGA_X_PKZB'}],
            'PL': [{'name': 'LDA_C_PZ'}],
            'P86': [{'name': 'GGA_C_P86'}],
            'B95': [{'name': 'MGGA_C_BC95'}],
            'KCIS': [{'name': 'MGGA_C_KCIS'}],
            'BRC': [{'name': 'MGGA_X_BR89'}],
            'VP86': [{'name': 'LDA_C_VWN_RPA'}, {'name': 'GGA_C_P86'}],
            'V5LYP': [{'name': 'LDA_C_VWN_RPA'}, {'name': 'GGA_C_LYP'}],
            'THCTH': [{'name': 'MGGA_X_TAU_HCTH'}, {'name': 'GGA_C_TAU_HCTH'}],
            'TPSSTPSS': [{'name': 'MGGA_C_TPSS'}, {'name': 'MGGA_X_TPSS'}],
            'B3LYP': [{'name': 'HYB_GGA_XC_B3LYP'}],
            'B3PW91': [{'name': 'HYB_GGA_XC_B3PW91'}],
            'B3P86': [{'name': 'HYB_GGA_XC_B3P86'}],
            'B1B95': [{'name': 'HYB_MGGA_XC_B88B95'}],
            'MPW1PW91': [{'name': 'HYB_GGA_XC_MPW1PW'}],
            'MPW1LYP': [{'name': 'HYB_GGA_XC_MPW1LYP'}],
            'MPW1PBE': [{'name': 'HYB_GGA_XC_MPW1PBE'}],
            'MPW3PBE': [{'name': 'HYB_GGA_XC_MPW3PBE'}],
            'B98': [{'name': 'HYB_GGA_XC_SB98_2C'}],
            'B971': [{'name': 'HYB_GGA_XC_B97_1'}],
            'B972': [{'name': 'HYB_GGA_XC_B97_2'}],
            'O3LYP': [{'name': 'HYB_GGA_XC_O3LYP'}],
            'TPSSH': [{'name': 'HYB_MGGA_XC_TPSSH'}],
            'BMK': [{'name': 'HYB_GGA_XC_B97_K'}],
            'X3LYP': [{'name': 'HYB_GGA_XC_X3LYP'}],
            'THCTHHYB': [{'name': 'HYB_MGGA_X_TAU_HCTH'}, {'name': 'GGA_C_HYB_TAU_HCTH'}],
            'BHANDH': [{'name': 'HYB_GGA_XC_BHANDH'}],
            'BHANDHLYP': [{'name': 'HYB_GGA_XC_BHANDHLYP'}],
            'APF': [{'name': 'HYB_GGA_XC_APF'}],
            'APFD': [{'name': 'HYB_GGA_XC_APFD'}],
            'HF': [{'name': 'HF_HF_X'}],
            'RHF': [{'name': 'HF_RHF_X'}],
            'UHF': [{'name': 'HF_UHF_X'}],
            'ROHF': [{'name': 'HF_ROHF_X'}],
            'CC': [{'name': 'HF_CCD'}],
            'QCID': [{'name': 'HF_CCD'}],
            'CCD': [{'name': 'HF_CCD'}],
            'CCSD': [{'name': 'HF_CCSD'}],
            'CCSD(T)': [{'name': 'HF_CCSD(T)'}],
            'CCSD-T': [{'name': 'HF_CCSD(T)'}],
            'CI': [{'name': 'HF_CI'}],
            'CID': [{'name': 'HF_CID'}],
            'CISD': [{'name': 'HF_CISD'}],
            'QCISD': [{'name': 'HF_QCISD'}],
            'QCISD(T)': [{'name': 'HF_QCISD(T)'}],
            'QCISD(TQ)': [{'name': 'HF_QCISD(TQ)'}],
            'OHSE2PBE': [{'name': 'HYB_GGA_XC_HSE03'}],
            'HSEH1PBE': [{'name': 'HYB_GGA_XC_HSE06'}],
            'OHSE1PBE': [{'name': 'HYB_GGA_XC_HSEOLD'}],
            'PBEH1PBE': [{'name': 'HYB_GGA_XC_PBEH1PBE'}],
            'PBE1PBE': [{'name': 'HYB_GGA_XC_PBEH'}],
            'M05': [{'name': 'MGGA_X_M05'}, {'name': 'MGGA_C_M05'}],
            'M052X': [{'name': 'HYB_MGGA_X_M05_2X'}, {'name': 'MGGA_C_M05_2X'}],
            'M06': [{'name': 'HYB_MGGA_X_M06'}, {'name': 'MGGA_C_M06'}],
            'M062X': [{'name': 'HYB_MGGA_X_M06_2X'}, {'name': 'MGGA_C_M06_2X'}],
            'M06HF': [{'name': 'HYB_MGGA_X_M06_HF'}, {'name': 'MGGA_C_M06_HF'}],
            'M11': [{'name': 'HYB_MGGA_X_M11'}, {'name': 'MGGA_C_M11'}],
            'MP2': [{'name': 'HF_MP2'}],
            'MP3': [{'name': 'HF_MP3'}],
            'MP4': [{'name': 'HF_MP4'}],
            'MP4(DQ)': [{'name': 'HF_MP4(DQ)'}],
            'MP4(SDQ)': [{'name': 'HF_MP4(SDQ)'}],
            'MP4(SDTQ)': [{'name': 'HF_MP4SDTQ'}],
            'MP5': [{'name': 'HF_MP5'}],
            'AM1': [{'name': 'HYB_AM1'}],
            'PM3': [{'name': 'HYB_PM3'}],
            'PM3MM': [{'name': 'HYB_PM3MM'}],
            'PM3D3': [{'name': 'HYB_PM3D3'}],
            'PM6': [{'name': 'HYB_PM6'}],
            'PM7': [{'name': 'HYB_PM7'}],
            'PM7R6': [{'name': 'HYB_PM7R6'}],
            'PM7MOPAC': [{'name': 'HYB_PM7MOPAC'}],
            'CBS-4': [{'name': 'HYB_CBS-4'}],
            'CBS-4M': [{'name': 'HYB_CBS-4M'}],
            'CBS-4O': [{'name': 'HYB_CBS-4O'}],
            'CBS-APNO': [{'name': 'HYB_CBS-APNO'}],
            'CBS-Q': [{'name': 'HYB_CBS-Q'}],
            'CBS-QB3': [{'name': 'HYB_CBS-QB3'}],
            'CBS-QB3O': [{'name': 'HYB_CBS-QB3O'}],
            'ROCBS-QB3': [{'name': 'HYB_ROCBS-QB3'}],
            'SOGGA11X': [{'name': 'HYB_GGA_X_SOGGA11_X'}, {'name': 'HYB_GGA_X_SOGGA11_X'}],
            'MN12SX': [{'name': 'HYB_MGGA_X_MN12_SX'}, {'name': 'MGGA_C_MN12_SX'}],
            'N12SX': [{'name': 'HYB_GGA_X_N12_SX'}, {'name': 'GGA_C_N12_SX'}],
            'LC-WPBE': [{'name': 'HYB_GGA_XC_LC_WPBE'}],
            'CAM-B3LYP': [{'name': 'HYB_GGA_XC_CAM_B3LYP'}],
            'WB97': [{'name': 'HYB_GGA_XC_WB97'}],
            'WB97X': [{'name': 'HYB_GGA_XC_WB97X'}],
            'WB97XD': [{'name': 'HYB_GGA_XC_WB97X_D'}],
            'HISSBPBE': [{'name': 'HYB_HISSBPBE'}],
            'B2PLYP': [{'name': 'HYB_B2PLYP'}],
            'MPW2PLYP': [{'name': 'HYB_MPW2PLYP'}],
            'B2PLYPD': [{'name': 'HYB_B2PLYPD'}],
            'MPW2PLYPD': [{'name': 'HYB_MPW2PLYPD'}],
            'B2PLYPD3': [{'name': 'HYB_B2PLYPD3'}],
            'MPW2PLYPD3': [{'name': 'HYB_MPW2PLYPD3'}],
            'G1': [{'name': 'HYB_G1'}],
            'G2': [{'name': 'HYB_G2'}],
            'G2MP2': [{'name': 'HYB_G2MP2'}],
            'G3': [{'name': 'HYB_G3'}],
            'G3B3': [{'name': 'HYB_G3B3'}],
            'G3MP2': [{'name': 'HYB_G3MP2'}],
            'G3MP2B3': [{'name': 'HYB_G3MP2B3'}],
            'G4': [{'name': 'HYB_G4'}],
            'G4MP2': [{'name': 'HYB_G4MP2'}],
            'LC-': [{'name': 'GGA_X_ITYH_LONG_RANGE'}],
            'BP': [{'name': 'GGA_X_B88'}, {'name': 'GGA_C_P86'}],
            'BP86': [{'name': 'GGA_X_B88'}, {'name': 'GGA_C_P86'}]
        }

        self._method_map = {
            'AMBER': [{'name': 'Amber'}],
            'DREIDING': [{'name': 'Dreiding'}],
            'UFF': [{'name': 'UFF'}],
            'AM1': [{'name': 'AM1'}],
            'PM3': [{'name': 'PM3'}],
            'PM3MM': [{'name': 'PM3MM'}],
            'PM3D3': [{'name': 'PM3D3'}],
            'PM6': [{'name': 'PM6'}],
            'PM7': [{'name': 'PM7'}],
            'PM7R6': [{'name': 'PM7R6'}],
            'PM7MOPAC': [{'name': 'PM7MOPAC'}],
            'PDDG': [{'name': 'PDDG'}],
            'CNDO': [{'name': 'CNDO'}],
            'INDO': [{'name': 'INDO'}],
            'MINDO': [{'name': 'MINDO'}],
            'MINDO3': [{'name': 'MINDO3'}],
            'ZINDO': [{'name': 'ZINDO'}],
            'HUCKEL': [{'name': 'HUCKEL'}],
            'EXTENDEDHUCKEL': [{'name': 'HUCKEL'}],
            'ONIOM': [{'name': 'ONIOM'}],
            'HF': [{'name': 'HF'}],
            'RHF': [{'name': 'RHF'}],
            'UHF': [{'name': 'UHF'}],
            'ROHF': [{'name': 'ROHF'}],
            'GVB': [{'name': 'GVB'}],
            'DFT': [{'name': 'DFT'}],
            'CI': [{'name': 'CI'}],
            'CID': [{'name': 'CID'}],
            'CISD': [{'name': 'CISD'}],
            'CIS': [{'name': 'CIS'}],
            'BD': [{'name': 'BD'}],
            'BD(T)': [{'name': 'BD(T)'}],
            'CC': [{'name': 'CCD'}],
            'QCID': [{'name': 'CCD'}],
            'CCD': [{'name': 'CCD'}],
            'CCSD': [{'name': 'CCSD'}],
            'EOMCCSD': [{'name': 'EOMCCSD'}],
            'QCISD': [{'name': 'QCISD'}],
            'CCSD(T)': [{'name': 'CCSD(T)'}],
            'CCSD-T': [{'name': 'CCSD(T)'}],
            'QCISD(T)': [{'name': 'QCISD(T)'}],
            'QCISD(TQ)': [{'name': 'QCISD(TQ)'}],
            'MP2': [{'name': 'MP2'}],
            'MP3': [{'name': 'MP3'}],
            'MP4': [{'name': 'MP4'}],
            'MP4DQ': [{'name': 'MP4DQ'}],
            'MP4(DQ)': [{'name': 'MP4DQ'}],
            'MP4SDQ': [{'name': 'MP4SDQ'}],
            'MP4(SDQ)': [{'name': 'MP4SDQ'}],
            'MP4SDTQ': [{'name': 'MP4SDTQ'}],
            'MP4(SDTQ)': [{'name': 'MP4SDTQ'}],
            'MP5': [{'name': 'MP5'}],
            'CAS': [{'name': 'CASSCF'}],
            'CASSCF': [{'name': 'CASSCF'}],
            'G1': [{'name': 'G1'}],
            'G2': [{'name': 'G2'}],
            'G2MP2': [{'name': 'G2MP2'}],
            'G3': [{'name': 'G3'}],
            'G3MP2': [{'name': 'G3MP2'}],
            'G3B3': [{'name': 'G3B3'}],
            'G3MP2B3': [{'name': 'G3MP2B3'}],
            'G4': [{'name': 'G4'}],
            'G4MP2': [{'name': 'G4MP2'}],
            'CBS-4': [{'name': 'CBS-4'}],
            'CBS-4M': [{'name': 'CBS-4M'}],
            'CBS-4O': [{'name': 'CBS-4O'}],
            'CBS-APNO': [{'name': 'CBS-APNO'}],
            'CBS-Q': [{'name': 'CBS-Q'}],
            'CBS-QB3': [{'name': 'CBS-QB3'}],
            'CBS-QB3O': [{'name': 'CBS-QB3O'}],
            'CBSEXTRAP': [{'name': 'CBSExtrapolate'}],
            'CBSEXTRAPOLATE': [{'name': 'CBSExtrapolate'}],
            'ROCBS-QB3': [{'name': 'ROCBS-QB3'}],
            'W1U': [{'name': 'W1U'}],
            'W1BD': [{'name': 'W1BD'}],
            'W1RO': [{'name': 'W1RO'}],
        }

        self._basis_set_map = {
            'STO-3G': [{'name': 'STO-3G'}],
            '3-21G': [{'name': '3-21G'}],
            '6-21G': [{'name': '6-21G'}],
            '4-31G': [{'name': '4-31G'}],
            '6-31G': [{'name': '6-31G'}],
            '6-311G': [{'name': '6-311G'}],
            'D95V': [{'name': 'D95V'}],
            'D95': [{'name': 'D95'}],
            'CC-PVDZ': [{'name': 'cc-pVDZ'}],
            'CC-PVTZ': [{'name': 'cc-pVTZ'}],
            'CC-PVQZ': [{'name': 'cc-pVQZ'}],
            'CC-PV5Z': [{'name': 'cc-pV5Z'}],
            'CC-PV6Z': [{'name': 'cc-pV6Z'}],
            'SV': [{'name': 'SV'}],
            'SVP': [{'name': 'SVP'}],
            'TZV': [{'name': 'TZV'}],
            'TZVP': [{'name': 'TZVP'}],
            'DEF2SV': [{'name': 'Def2SV'}],
            'DEF2SVP': [{'name': 'Def2SVP'}],
            'DEF2SVPP': [{'name': 'Def2SVPP'}],
            'DEF2TZV': [{'name': 'Def2TZV'}],
            'DEF2TZVP': [{'name': 'Def2TZVP'}],
            'DEF2TZVPP': [{'name': 'Def2TZVPP'}],
            'DEF2QZV': [{'name': 'Def2QZV'}],
            'DEF2QZVP': [{'name': 'Def2QZVP'}],
            'DEF2QZVPP': [{'name': 'Def2QZVPP'}],
            'QZVP': [{'name': 'QZVP'}],
            'MIDIX': [{'name': 'MidiX'}],
            'EPR-II': [{'name': 'EPR-II'}],
            'EPR-III': [{'name': 'EPR-III'}],
            'UGBS': [{'name': 'UGBS'}],
            'MTSMALL': [{'name': 'MTSmall'}],
            'DGDZVP': [{'name': 'DGDZVP'}],
            'DGDZVP2': [{'name': 'DGDZVP2'}],
            'DGTZVP': [{'name': 'DGTZVP'}],
            'CBSB3': [{'name': 'CBSB3'}],
            'CBSB7': [{'name': 'CBSB7'}],
            'SHC': [{'name': 'SHC'}],
            'SEC': [{'name': 'SHC'}],
            'CEP-4G': [{'name': 'CEP-4G'}],
            'CEP-31G': [{'name': 'CEP-31G'}],
            'CEP-121G': [{'name': 'CEP-121G'}],
            'LANL1': [{'name': 'LANL1'}],
            'LANL2': [{'name': 'LANL2'}],
            'SDD': [{'name': 'SDD'}],
            'OLDSDD': [{'name': 'OldSDD'}],
            'SDDALL': [{'name': 'SDDAll'}],
            'GEN': [{'name': 'General'}],
            'GENECP': [{'name': 'General ECP'}],
            'CHKBAS': [{'name': 'CHKBAS'}],
            'EXTRABASIS': [{'name': 'ExtraBasis'}],
            'DGA1': [{'name': 'DGA1'}],
            'DGA2': [{'name': 'DGA2'}],
            'SVPFIT': [{'name': 'SVPFit'}],
            'TZVPFIT': [{'name': 'TZVPFit'}],
            'W06': [{'name': 'W06'}],
            'CHF': [{'name': 'CHF'}],
            'FIT': [{'name': 'FIT'}],
            'AUTO': [{'name': 'AUTO'}],
        }

        self._xc_functional_pattern = re.compile(
            r'(?:#?[Pp]?)(?:(WPBEH)(.*)|(LC\-)(.*)|(G96)(.*)|(BRX)(.*)|(SV)(.*)|(B)([^2^3]*)|(O)([^2^3]))')

        self._basis_set_pattern = re.compile(
            r'(6\-311G)|(6\-31G)|(LANL2)|(LANL1)|(CBSB7)|(UGBS)|AUG\-(.*)|(D95)')

        self._energy_methods = {
            'mp': (x_gaussian_section_moller_plesset, [
                'mp2', 'mp3', 'mp4dq', 'mp4sdq', 'mp4sdtq', 'mp5']),
            'cc': (x_gaussian_section_coupled_cluster, ['ccsd']),
            'qci': (x_gaussian_section_quadratic_ci, ['qcisd', 'qcisdtq']),
            'ci': (x_gaussian_section_ci, ['ci']),
            'semiempirical': (x_gaussian_section_semiempirical, ['']),
            'molmech': (x_gaussian_section_molmech, [])}

    def parse_scc(self, section):
        sec_run = self.archive.section_run[-1]
        sec_scc = sec_run.m_create(SingleConfigurationCalculation)

        # total energy
        energy_total = section.get('energy_total')
        if energy_total is not None:
            sec_scc.energy_total = energy_total[-1]

        # hybrid coefficients
        sec_hybrid_coeffs = None
        for key in ['coeff1', 'coeff2']:
            val = section.get('hybrid_xc_%s' % key)
            if val is None:
                continue
            if sec_hybrid_coeffs is None:
                sec_hybrid_coeffs = sec_scc.m_create(x_gaussian_section_hybrid_coeffs)
            setattr(sec_hybrid_coeffs, 'hybrid_xc_%s' % key, val)

        def parse_energy_corrections(method, iteration=False):

            if not iteration and section.get(method) is None:
                return

            sec_method, energy_keys = self._energy_methods.get(method, (None, []))
            if sec_method is None:
                return

            sec_method = sec_scc.m_create(sec_method)
            for key in energy_keys:
                key = '%s_correction_energy' % key
                val = section.get(key)
                if val is not None:
                    setattr(sec_method, 'x_gaussian_%s' % key, val[-1])

            energy_method = section.get('%s_method' % method)
            if energy_method is not None:
                setattr(sec_method, 'x_gaussian_%s_method' % method, energy_method)

            energy = section.get('%s_energy' % method)
            if energy is not None:
                setattr(sec_method, 'x_gaussian_%s_energy' % method, energy)

            energy_total = section.get('energy_total_%s' % method)
            if energy_total is not None:
                sec_scc.energy_total = energy_total[-1]

            return sec_method

        # energy corrections calculated from different methods
        sec_energy_method = None
        for method in ['mp', 'cc', 'qci', 'ci', 'semiempirical', 'molmech']:
            sec_energy_method = parse_energy_corrections(method)
            if sec_energy_method is not None:
                break

        # TODO make metainfo accept array not separate sections
        # excited state
        excited_state = section.get('excited_state', [])
        for state in excited_state:
            sec_excited_state = sec_scc.m_create(x_gaussian_section_excited)
            sec_excited_state.x_gaussian_excited_state_number = int(state[0])
            sec_excited_state.x_gaussian_excited_energy = float(state[1]) * ureg.eV
            sec_excited_state.x_gaussian_excited_oscstrength = float(state[2])
            sec_excited_state.x_gaussian_excited_spin_squared = float(state[3])
            sec_excited_state.x_gaussian_excited_transition = ' '.join(state[4:])

        # casscf
        casscf_energy = section.get('casscf_energy', [])
        for energy in casscf_energy:
            sec_casscf = sec_scc.m_create(x_gaussian_section_casscf)
            sec_casscf.x_gaussian_casscf_energy = energy

        # orbital symmetries
        orbital_symmetries = section.get('population_analysis', {}).get('orbital_symmetries')
        if orbital_symmetries is not None:
            sec_orbital_symmetries = sec_run.m_create(x_gaussian_section_orbital_symmetries)
            if orbital_symmetries[:2]:
                sec_orbital_symmetries.x_gaussian_alpha_symmetries = np.concatenate(orbital_symmetries[:2])
            if orbital_symmetries[2:]:
                sec_orbital_symmetries.x_gaussian_beta_symmetries = np.concatenate(orbital_symmetries[2:])

        # electronic symmetries
        elstate_symmetry = section.get('population_analysis', {}).get('x_gaussian_elstate_symmetry')
        if elstate_symmetry is not None:
            sec_symmetry = sec_run.m_create(x_gaussian_section_symmetry)
            sec_symmetry.x_gaussian_elstate_symmetry = elstate_symmetry

        # eigenvalues
        values, occupation = [[], []], [[], []]
        eigenvalues = section.get('population_analysis', {}).get('eigenvalues')
        if eigenvalues is not None:
            for eigs_occs in eigenvalues:
                for spin in range(2):
                    values[spin].extend(eigs_occs[0][spin])
                    occupation[spin].extend(eigs_occs[1][spin])
            if not values[1]:
                values = values[0:1]
                occupation = np.array(occupation[0:1]) * 2
            try:
                values = np.reshape(values, (len(values), 1, len(values[0])))
                occupation = np.reshape(occupation, (len(occupation), 1, len(occupation[0])))
                sec_eigenvalues = sec_scc.m_create(Eigenvalues)
                sec_eigenvalues.eigenvalues_values = np.array(values, dtype=np.dtype(np.float64)) * ureg.hartree
                sec_eigenvalues.eigenvalues_occupation = occupation
            except Exception:
                self.logger.error('Error setting eigenvalues.')

        # optimization
        optimization_completed = section.get('optimization_completed')
        if optimization_completed is not None:
            sec_optimization_info = sec_run.m_create(x_gaussian_section_geometry_optimization_info)
            sec_optimization_info.x_gaussian_geometry_optimization_converged = optimization_completed

        # multipoles
        multipoles = []
        keys = ['charge', 'dipole', 'quadrupole', 'octapole', 'hexadecapole']
        for n, key in enumerate(keys):
            val = section.get(key)
            if val is not None:
                val = val.to('coulomb * meter**%d' % n).magnitude
                val = [val] if isinstance(val, float) else val
                # TODO why quadrupole moments are not sorted, adapted from old parser
                if key == 'quadrupole':
                    val = [val[0], val[3], val[1], val[4], val[5], val[2]]
                multipoles.extend(val)
        if multipoles:
            sec_molecular_multipoles = sec_scc.m_create(x_gaussian_section_molecular_multipoles)
            sec_molecular_multipoles.x_gaussian_molecular_multipole_values = multipoles
            sec_molecular_multipoles.x_gaussian_molecular_multipole_m_kind = 'polynomial'

        # forces
        forces = section.get('forces')
        if forces is not None:
            sec_scc.atom_forces_raw = forces

        # force constants
        force_constants = section.get('force_constants')
        if force_constants is not None:
            sec_force_constant = sec_run.m_create(x_gaussian_section_force_constant_matrix)
            force_constants = force_constants * ureg.hartree / ureg.bohr ** 2
            sec_force_constant.x_gaussian_force_constant_values = force_constants.to(
                'J/(m**2)').magnitude

        # vibrational frequencies
        frequencies = section.get('frequencies')
        if frequencies is not None:
            # frequencies in old parsers are in J, not consistent with metainfo
            sec_frequencies = sec_run.m_create(x_gaussian_section_frequencies)
            sec_frequencies.x_gaussian_frequencies = np.hstack(frequencies)
            reduced_masses = section.get('reduced_masses')
            if reduced_masses is not None:
                reduced_masses = np.hstack(reduced_masses) * ureg.amu
                sec_frequencies.x_gaussian_red_masses = reduced_masses.to('kg').magnitude
            normal_modes = section.get('normal_modes')
            if normal_modes is not None:
                normal_modes = np.hstack(normal_modes)
                normal_modes = np.reshape(normal_modes, (len(normal_modes), len(normal_modes[0]) // 3, 3))
                sec_frequencies.x_gaussian_normal_mode_values = np.transpose(normal_modes, axes=(1, 0, 2))

        # thermochemistry
        temperature_pressure = section.get('temperature_pressure')
        if temperature_pressure is not None:
            sec_thermochem = sec_run.m_create(x_gaussian_section_thermochem)
            sec_thermochem.x_gaussian_temperature = temperature_pressure[0]
            sec_thermochem.x_gaussian_pressure = (temperature_pressure[1] * ureg.atm).to('N/m**2').magnitude
            moments = section.get('moments')
            if moments is not None:
                sec_thermochem.x_gaussian_moments = moments.to('kg*m**2').magnitude
            keys = [
                'zero_point_energy', 'thermal_correction_energy',
                'thermal_correction_enthalpy', 'thermal_correction_free_energy']
            for key in keys:
                val = section.get(key)
                if val is not None:
                    setattr(sec_thermochem, 'x_gaussian_%s' % key, val.to('joule').magnitude)

        # scf_iteration
        for iteration in section.get('scf_iteration', []):
            sec_scf_iteration = sec_scc.m_create(ScfIteration)
            for key in ['energy_total_scf_iteration', 'x_gaussian_delta_energy_total_scf_iteration']:
                val = iteration.get(key)
                if val is not None:
                    setattr(sec_scf_iteration, key, val)
        iteration = section.get('scf_iteration_final')
        if iteration is not None:
            sec_scf_iteration = sec_scc.section_scf_iteration
            sec_scf_iteration = sec_scf_iteration[-1] if sec_scf_iteration else sec_scc.m_create(ScfIteration)
            keys = [
                'x_gaussian_single_configuration_calculation_converged',
                'x_gaussian_hf_detect', 'x_gaussian_energy_scf', 'x_gaussian_energy_error',
                'x_gaussian_electronic_kinetic_energy'
                'x_gaussian_spin_S2', 'x_gaussian_after_annihilation_spin_S2',
                'x_gaussian_perturbation_energy']
            for key in keys:
                val = iteration.get(key)
                if val is not None:
                    setattr(sec_scf_iteration, key, val)
            if iteration.get('x_gaussian_single_configuration_calculation_converged') is not None:
                energy_scf = iteration.get('x_gaussian_energy_scf')
                electronic_ke = iteration.get('x_gaussian_electronic_kinetic_energy')
                if energy_scf is not None and electronic_ke is not None:
                    sec_scf_iteration.x_gaussian_energy_electrostatic = energy_scf - electronic_ke
                if energy_scf is not None:
                    sec_scc.energy_total = energy_scf
                sec_scc.single_configuration_calculation_converged = True

        return sec_scc

    def parse_system(self, n_run, section):
        sec_run = self.archive.section_run[-1]
        sec_system = sec_run.m_create(System)
        lattice_vector = self.out_parser.get('run')[n_run].get('lattice_vector')
        if lattice_vector is not None:
            sec_system.lattice_vectors = lattice_vector * ureg.angstrom
            sec_system.simulation_cell = lattice_vector * ureg.angstrom
            pbc = [len(lattice_vector) >= n for n in [1, 2, 3]]
            sec_system.configuration_periodic_dimensions = [pbc]
        else:
            sec_system.lattice_vectors = np.zeros((3, 3))
            sec_system.configuration_periodic_dimensions = [False] * 3

        orientation = section.get('standard_orientation')
        if orientation is None:
            orientation = section.get('input_orientation')
        if orientation is None:
            orientation = section.get('z_matrix_orientation')
        if orientation is not None:
            sec_system.atom_labels = [
                ase.data.chemical_symbols[int(n)] for n in [o[1] for o in orientation]]
            sec_system.atom_positions = [o[-3:] for o in orientation] * ureg.angstrom
            sec_system.x_gaussian_number_of_atoms = len(sec_system.atom_labels)

        for key in ['x_gaussian_total_charge', 'x_gaussian_spin_target_multiplicity']:
            val = self.out_parser.get('run')[n_run].get(key)
            if val is not None:
                setattr(sec_system, key, val)

        return sec_system

    def parse_configurations(self, n_run):
        systems = self.out_parser.get('run')[n_run].get('system', [])
        for system in systems:
            sec_system = self.parse_system(n_run, system)
            for calculation in system.get('calculation', [system]):
                sec_scc = self.parse_scc(calculation)
                if sec_scc is not None:
                    if sec_system is not None:
                        sec_scc.single_configuration_calculation_to_system_ref = sec_system
                    sec_method = self.archive.section_run[-1].section_method
                    if sec_method:
                        sec_scc.single_configuration_to_calculation_method_ref = sec_method[-1]

        iterations = self.out_parser.get('run')[n_run].get('iteration', [])
        sec_system = self.archive.section_run[0].section_system
        sec_method = self.archive.section_run[0].section_method
        for iteration in iterations:
            sec_scc = self.parse_scc(iteration)
            if sec_scc is not None:
                if sec_system:
                    sec_scc.single_configuration_calculation_to_system_ref = sec_system[-1]
                if sec_method:
                    sec_scc.single_configuration_to_calculation_method_ref = sec_method[-1]

    def parse_method(self, n_run):
        sec_run = self.archive.section_run[-1]
        sec_method = sec_run.m_create(Method)
        run = self.out_parser.get('run')[n_run]

        for key in ['x_gaussian_settings_corrected']:
            val = run.get(key)
            if val is None:
                continue
            setattr(sec_method, key, val)

        def resolve_method(parameter):
            def get_method(name):
                method = self._method_map.get(parameter, None)
                if method is not None:
                    return parameter

                method = self._xc_functional_map.get(parameter, None)
                return 'DFT' if method is not None else method

            method = get_method(parameter)
            if method is None:
                # TODO This is inefficient. Adapted from old parser
                for n in range(9, 1, -1):
                    method = get_method(parameter[0:n])
                    if method is not None:
                        break

            return method

        def resolve_prefix(name):
            name = name.split('=')[0].split('-')[0].strip()
            prefix = ''
            if name not in ['RHF', 'ROHF', 'UHF', 'UFF']:
                res = re.match(r'([RU]{1}O?)(.*)', name)
                if res is not None:
                    prefix = res[1] if res[1] else ''
                    name = res[2]
            return prefix, name

        def resolve_basis_set(parameter):
            basis_set = self._basis_set_map.get(parameter, None)
            if basis_set is not None:
                return (parameter, parameter)

            res = self._basis_set_pattern.match(parameter)
            if res is not None:
                basis_keys = [key for key in res.groups() if key is not None]
                if len(basis_keys) != 1:
                    self.logger.warn('Cannot resolve basis set', data=dict(key=parameter))
                return (basis_keys[0], parameter)

        def resolve_xc_functional(parameter):
            xc_functional = self._xc_functional_map.get(parameter, None)
            if xc_functional is not None:
                return parameter

            res = self._xc_functional_pattern.match(parameter)
            if res is not None:
                xc_keys = [key for key in res.groups() if key is not None]
                if len(xc_keys) != 2:
                    self.logger.warn('Cannot resolve xc functional', data=dict(key=name))
                    return

                x_name = 'LC-' if xc_keys[0] == 'LC-' else self._xc_functional_map.get(
                    xc_keys[0], [{}])[0].get('name', '')
                c_index = -1 if xc_keys[0] == 'LC-' else 0
                c_name = self._xc_functional_map.get(xc_keys[1], [{}])[c_index].get('name', '')

                if x_name and c_name:
                    return '%s%s' % (x_name, c_name)

        settings = run.get('x_gaussian_settings_corrected', '').upper()
        settings = re.sub(r'(?:#[Pp]? |# *)', '', settings)
        xc_functionals = set()
        basis_sets = set()
        methods = set()
        prefix = ''
        for setting in settings.split():
            parameter = setting.split('/')
            prefix, name = resolve_prefix(parameter[0])

            method = resolve_method(name)
            if method is not None:
                methods.add(method)

            xc_functional = resolve_xc_functional(name)
            if xc_functional is not None:
                xc_functionals.add(xc_functional)

            basis_set_parameter = parameter[0] if not parameter[1:] else parameter[1]
            basis_set = resolve_basis_set(basis_set_parameter.strip())
            if basis_set is not None:
                basis_sets.add(basis_set)

        if len(methods) != 1:
            self.logger.error('Found mutiple or no method', data=dict(
                n_parsed=len(methods)))
        for method in methods:
            sec_method.x_gaussian_method = method
            sec_method.electronic_structure_method = method
            for entry in self._method_map.get(method, []):
                sec_elstruc_method = sec_method.m_create(x_gaussian_section_elstruc_method)
                sec_elstruc_method.x_gaussian_electronic_structure_method = '%s%s' % (
                    prefix, entry['name'])

        if len(basis_sets) != 1:
            self.logger.error('Found multiple or no basis set', data=dict(
                n_parsed=len(basis_sets)))
        for basis_set in basis_sets:
            sec_method.basis_set = basis_set[0]
            for _ in self._basis_set_map.get(basis_set[0], []):
                sec_basis_set_atom_centered = sec_run.m_create(BasisSetAtomCentered)
                # old parser writes the full name of basis set here not the name on map
                sec_basis_set_atom_centered.basis_set_atom_centered_short_name = basis_set[1]

        if len(xc_functionals) != 1:
            self.logger.error('Found multiple or no xc functional', data=dict(
                n_parsed=len(xc_functionals)))
        for xc_functional in xc_functionals:
            sec_method.x_gaussian_xc = xc_functional
            for entry in self._xc_functional_map.get(xc_functional, []):
                sec_xc_functionals = sec_method.m_create(XCFunctionals)
                sec_xc_functionals.XC_functional_name = entry['name']

        sec_sampling_method = sec_run.m_create(SamplingMethod)
        sec_sampling_method.sampling_method = 'geometry_optimization'

    def init_parser(self, filepath, logger):
        self.out_parser.mainfile = filepath
        self.out_parser.logger = logger

    def parse(self, filepath, archive, logger):
        self.filepath = filepath
        self.archive = archive
        self.logger = logging.getLogger(__name__) if logger is None else logger
        self.init_parser(filepath, logger)

        runs = self.out_parser.get('run', [])
        for n in range(len(runs)):
            sec_run = self.archive.m_create(Run)
            sec_run.program_name = 'Gaussian'
            sec_run.program_basis_set_type = 'gaussians'

            program = self.out_parser.get('program')
            sec_run.program_version = program[0]
            sec_run.x_gaussian_program_implementation = program[1]
            sec_run.x_gaussian_program_release_date = program[2]
            sec_run.x_gaussian_program_execution_date = program[3]

            for key in ['x_gaussian_chk_file', 'x_gaussian_memory', 'x_gaussian_number_of_processors']:
                val = self.out_parser.get(key)
                if val is not None:
                    setattr(sec_run, key, val)

            self.parse_method(n)

            self.parse_configurations(n)

            sec_times = sec_run.m_create(x_gaussian_section_times)
            for key in ['program_cpu_time', 'program_termination_date']:
                val = runs[n].get(key)
                if val is not None:
                    setattr(sec_times, 'x_gaussian_%s' % key, val.strip())
